/*-------------------------------------------------------------------*/
/* QMODEL.C      Pavel Kopylov   (justas@cs.joensuu.fi)              */
/*               Eugene Ageenko                                      */
/*                                                                   */
/* Q modelling                                                       */
/*-------------------------------------------------------------------*/

#define  ProgName       "QMODEL"
#define  VersionNumber  "Version L.02"
#define  LastUpdated    "21.09.2000"

#include "../Modules/owntypes.h"
#include "qmodel.h"
#include "../Modules/memctrl.h"
#include "../Modules/file.h"
#include "../Modules/genio.h"

// Private definitions
__inline int FindNearestFastAttackState(GENMODEL *, double);



/* ---------------------------------------------------------- */

YESNO InitQModel(GENMODEL* pGModel, ULONG NModels, int NSymbols, int Process, int ModellingType)
{
  pGModel->NModels       = NModels;
  pGModel->ModellingType = (MODELLINGTYPE) ModellingType;
  pGModel->NSymbols      = NSymbols;
  pGModel->Process       = (PROCESSTYPE) Process;

  pGModel->Model = allocate(sizeof(QMODEL)*NModels);
  if (pGModel->Model == NULL) return(NO);

  memset(pGModel->Model, 0, sizeof(QMODEL)*NModels);

  return (YES);
}

/* ---------------------------------------------------------- */

void ReInitQModel(GENMODEL* pGModel)
{
  ULONG i; 
  QMODEL *QModel = (QMODEL*)pGModel->Model;

  switch (pGModel->ModellingType)
  {                                                         
    case FORWARDADAPTIVE:
         for (i=0; i < pGModel->NModels; i++)     
         {
           QModel[i].MPS   = (QModel[i].FAIndex & 0x80) ? 1 : 0;
           QModel[i].Index = (QModel[i].FAIndex & 0x7f);
         }
         break;
    case ADAPTIVE:
    default:
         for(i=0; i < pGModel->NModels; i++)
         {
           QModel[i].MPS = 0;
           QModel[i].Index = 0;
         }
         break;
  }
}

/* ---------------------------------------------------------- */

void DoneQModel(GENMODEL* pGModel)
{       
  deallocate(pGModel->Model);
  pGModel->Model=NULL;
}


/*==================== Forward-adaptive Q-based coders ================*/

// return YES = ok, NO = not ok
int StoreInitialQModel(GENMODEL *pGModel, FLOW *f)
{
  ULONG i;
  QMODEL *QModel = (QMODEL*)pGModel->Model;

  //Arithmetic output
  WRITEUINT(f, pGModel->NModels, 4);

  for (i=0; i < pGModel->NModels; i++)
  {
    f->OutputByte(f,QModel[i].FAIndex);
  }                         

  if (f->Status) return YES; 
  else           return NO;
}

/* ---------------------------------------------------------- */

// return YES = ok, NO = not ok, NOEX (-1) if model file does not fit
int LoadInitialQModel(GENMODEL *pGModel, FLOW *f)
{
  ULONG i=0;
  ULONG NModels;
  ULONG tmp_ulong;
  QMODEL *QModel = (QMODEL*)pGModel->Model;
                            
  //Arithmetic input
  READUINT(f, NModels, 4);
  if (pGModel->NModels != NModels) return NOEX;

  if (pGModel->ModellingType != ADAPTIVE)
  {
    for (i=0; i<pGModel->NModels; i++)
    {
      QModel[i].FAIndex = f->InputByte(f);
      if (!(f->Status)) return NO;
    }
  }
  return(YES); // model is read
}

/* ---------------------------------------------------------- */

void CreateInitialQModel(GENMODEL *pGModel, ULONG n_total[], ULONG n_whites[])
{
  ULONG i;
  int index;
  double  WhiteProb, LpsProb;
  QMODEL *QModel = (QMODEL*)pGModel->Model;

  for(i=0; i < pGModel->NModels; i++)
  {                      
    if (n_total[i] == 0) QModel[i].FAIndex = 0;
    else 
    {
      WhiteProb = (double)n_whites[i]/n_total[i];

      LpsProb   = WhiteProb < 0.5 ? WhiteProb : 1-WhiteProb;
      index     = WhiteProb < 0.5 ? 0x80 : 0x00;
      index     = index | FindNearestFastAttackState(pGModel, LpsProb);
      QModel[i].FAIndex = index;
    }
  }
}

/* ---------------------------------------------------------- */

__inline int FindNearestFastAttackState(GENMODEL *pGModel, double Prob)
{
  int    i;

  if (Prob == 0) return 13;

  for(i=0; i<13; i++)
    if(Prob > pGModel->FastAttack[i]) break;
  
  return(pGModel->FastAttackStateNumber[i]); // i - if terminated, 13 if no one found
}

