/*-------------------------------------------------------------------*/
/* QSMODEL.C     Pavel Kopylov   (justas@cs.joensuu.fi)              */
/*               Eugene Ageenko                                      */
/*                                                                   */
/* Quasi-static model (Range coder)                                  */
/*-------------------------------------------------------------------*/

#define  ProgName       "QSMODEL"
#define  VersionNumber  "Version L.01"
#define  LastUpdated    "21.09.2000"

#include <stdlib.h>  // exit()
#include "qsmodel.h"
#include "../Modules/memctrl.h"
#include "../Modules/interfc.h"
#include "../Modules/genio.h"


// Private definitions
void initqsmodel(QSMODEL *m, int n, int lg_totf, int rescale, PROCESSTYPE process);


/* ---------------------------------------------------------- */

YESNO InitQSModel(GENMODEL *pGModel, ULONG NModels, int NSymbols, int Process, int ModellingType)
{
  ULONG i;
  int k, end, initval;
  QSMODEL *m;

   
  pGModel->NModels       = NModels;
  pGModel->ModellingType = (MODELLINGTYPE) ModellingType;
  pGModel->NSymbols      = NSymbols;
  pGModel->Process       = (PROCESSTYPE) Process;

  pGModel->Model = allocate(sizeof(QSMODEL)*NModels);

  if (pGModel->Model==NULL) return NO;

  m = (QSMODEL*) pGModel->Model;
  for(i=0;i<NModels;i++) 
    initqsmodel(m+i, NSymbols, BITSINCNT, (1<<(BITSINCNT-1))-10, pGModel->Process);

  for(i=0;i<NModels;i++)
  {
     if (ModellingType != ADAPTIVE)
    {
      (m+i)->init = allocate(sizeof(UINT16)*NSymbols);
      memset(m+i, 0, sizeof(UINT16)*NSymbols);
    }
    else 
      (m+i)->init = NULL;
  }

  
  

//	ReInitQSModel(pGModel);
// DO NOT CALL this here, since model is not yet loaded in !ADAPTIVE case

  for(i=0; i<NModels; i++) // i run for number of models
  {
    m->rescale = m->n>>4 | 2;
    m->nextleft = 0;
    initval = m->cf[m->n] / m->n;
    end = m->cf[m->n] % m->n;
    for (k=0; k<end; k++) m->newf[k] = initval+1;
    for (; k<m->n; k++) m->newf[k] = initval;
    dorescale(m);
    m++;
  }
  
  return YES;
}

/* ---------------------------------------------------------- */

void DoneQSModel(GENMODEL *pGModel)
{
  ULONG i;
  QSMODEL *m = pGModel->Model;

  for(i=0;i<pGModel->NModels;i++)
  {
    deallocate(m->cf);         
    deallocate(m->newf);
    if (m->search != NULL) deallocate(m->search);
    if (m->init   != NULL) deallocate(m->init);
    m++;  // instead of running at (m+i)-> we increment just m!
  }
  deallocate(pGModel->Model);
  pGModel->Model=NULL;

}

/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */

void dorescale(QSMODEL *m)
{
  int i, cf, missing;
  int tmp;
  int start, end;

  if (m->nextleft)                    // we have some more before actual rescaling
  {
    m->incr++;
    m->left = m->nextleft;
    m->nextleft = 0;
    return;
  }
  if (m->rescale < m->targetrescale)  // double rescale interval if needed
  {
    m->rescale <<= 1;
    if (m->rescale > m->targetrescale) m->rescale = m->targetrescale;
  }
  cf = missing = m->cf[m->n];  /* do actual rescaling */

  for(i=m->n-1; i; i--)
  {
    tmp = m->newf[i];
    cf -= tmp;
    m->cf[i] = cf;
    tmp = tmp>>1 | 1;
    missing -= tmp;
    m->newf[i] = tmp;
  }
  if (cf!=m->newf[0])
  {
    ErrorMessage("QS model: rescaling left %d total frequency\n",cf);
    deallocate(m->cf);         
    deallocate(m->newf);
    if (m->search != NULL) deallocate(m->search);
    exit(1);
  }
  m->newf[0] = m->newf[0]>>1 | 1;
  missing -= m->newf[0];
  m->incr = missing / m->rescale;
  m->nextleft = missing % m->rescale;
  m->left = m->rescale - m->nextleft;

  if (m->search != NULL)
  {
    i=m->n;
    while (i)
    { 
      end = (m->cf[i]-1) >> m->searchshift;
      i--;
      start = m->cf[i] >> m->searchshift;
      while (start<=end)
      {
        m->search[start] = i;
        start++;
      }
    }
  }
}

/* ---------------------------------------------------------- */

// initialisation of qsmodel
// m   qsmodel to be initialized
// n   number of symbols in that model
// lg_totf  base2 log of total frequency count
// rescale  desired rescaling interval, should be < 1<<(lg_totf+1)
// init  array of int's to be used for initialisation (NULL ok)
// compress  set to 1 on compression, 0 on decompression
void initqsmodel(QSMODEL *m, int n, int lg_totf, int rescale, PROCESSTYPE process)
{
  m->n = n;
  m->targetrescale = rescale;
  m->searchshift = lg_totf - TBLSHIFT;
  if (m->searchshift < 0) m->searchshift = 0;
  m->cf = allocate((n+1)*sizeof(UINT16));
  m->newf = allocate((n+1)*sizeof(UINT16));
  m->cf[n] = 1<<lg_totf;
  m->cf[0] = 0;
  if (process==ENCODING) m->search = NULL;
  else
  {
    m->search = allocate(((1<<TBLSHIFT)+1)*sizeof(UINT16));
    m->search[1<<TBLSHIFT] = n-1;
  }
}

/* ---------------------------------------------------------- */

// reinitialisation of qsmodel
// m   qsmodel to be initialized
// init  array of int's to be used for initialisation (NULL ok)
void ReInitQSModel(GENMODEL *pGModel)
{   
  int k, end, initval;
  ULONG i;
  QSMODEL *QSModel = pGModel->Model;


  for(i=0; i < pGModel->NModels; i++) // i run for number of models
  {
      QSModel->rescale = QSModel->n>>4 | 2;
      QSModel->nextleft = 0;
      switch (pGModel->ModellingType)
      {                                                         
        case FORWARDADAPTIVE:
        // k run for array of symbols
            if (QSModel->init != NULL)
            {
              for(k=0; k<QSModel->n; k++) QSModel->newf[k] = QSModel->init[k];
              break;
            }
            // else we continue as in ADAPTIVE case

        case ADAPTIVE:
        default:
            initval = QSModel->cf[QSModel->n] / QSModel->n;
            end = QSModel->cf[QSModel->n] % QSModel->n;
            for (k=0; k<end; k++) QSModel->newf[k] = initval+1;
            for (; k<QSModel->n; k++) QSModel->newf[k] = initval;
            break;
      }
      dorescale(QSModel);
      QSModel++;
  }
}


void CreateInitialQSModel(GENMODEL *pGModel, ULONG *n_total, ULONG *n_whites)
{
  int k;
  ULONG i;
  QSMODEL *QSModel = pGModel->Model;
  
  for(i=0; i < pGModel->NModels; i++) // i run for number of models
  {
     if (QSModel->init == NULL) return;
     
     for(k=0; k<QSModel->n; k++) 
         QSModel->init[k] = 0;  // we do not know yet what to put in here ?
  }

}

int LoadInitialQSModel(GENMODEL *pGModel, FLOW *f)
{
  return 0;  
}

int StoreInitialQSModel(GENMODEL *pGModel, FLOW *f)
{
  return 1;
}


