/*-------------------------------------------------------------------*/
/* RNGCODER.C                                                        */
/*                                                                   */
/* Range-coder module;                                               */
/*-------------------------------------------------------------------*/
#define  ModuleName     "RNGCODER"
#define  VersionNumber  "Version 0.01"
#define  LastUpdated    "22.09.2000"

#include <stdlib.h>   // exit()
#include "../Modules/owntypes.h"
#include "../Modules/interfc.h"
#include "../Modules/memctrl.h"
#include "rngcoder.h"
#include "../Modules/genio.h"

#define CODE_BITS 32
#define Top_value ((CODE_VALUE)1 << (CODE_BITS-1))

#define SHIFT_BITS (CODE_BITS - 9)
#define EXTRA_BITS ((CODE_BITS-2) % 8 + 1)
#define Bottom_value (Top_value >> 8)

#define ESC         0xff        // escape       
#define STUFF       0x00        // escape escape
#define SDNORM      0x02        // Normal end

// Private definitions
void InitEnc(RCENC *pRCoder, char c, int initlength);
int  InitDec(RCENC *pRCoder);
void encode_freq(RCENC *pRCoder, CODE_VALUE sy_f, CODE_VALUE lt_f, CODE_VALUE tot_f );
void encode_shift(RCENC *pRCoder, CODE_VALUE sy_f, CODE_VALUE lt_f);
#define encode_byte(ac,b)  encode_shift(ac,(CODE_VALUE)1,(CODE_VALUE)(b),(CODE_VALUE)8)
#define encode_short(ac,s) encode_shift(ac,(CODE_VALUE)1,(CODE_VALUE)(s),(CODE_VALUE)16)
CODE_VALUE decode_culfreq(RCENC *pRCoder, CODE_VALUE tot_f );
CODE_VALUE decode_culshift(RCENC *ac, CODE_VALUE shift);
void decode_update(RCENC *pRCoder, CODE_VALUE sy_f, CODE_VALUE lt_f, CODE_VALUE tot_f);
#define decode_update_shift(pRCoder,f1,f2,f3) decode_update((pRCoder),(f1),(f2),(CODE_VALUE)1<<(f3));
unsigned char decode_byte(RCENC *pRCoder);
unsigned short decode_short(RCENC *pRCoder);

void qsgetfreq(QSMODEL *m, int sym, int *sy_f, int *lt_f );
int qsgetsym(QSMODEL *m, int lt_f );
void qsupdate(QSMODEL *m, int sym );



/*=========================== User part =============================*/

int RInputByte(RCENC *pRCoder)
{
  int x;

  if (pRCoder->pacfeed == 1 ) return(0);
  x = pRCoder->fByteIn(pRCoder->pCodeStream);
  if(x != ESC) 
  {
    pRCoder->BytesCount++;
    return (x);
  }
 
  x = pRCoder->fByteIn(pRCoder->pCodeStream);
  switch(x)
  {
    case STUFF:  pRCoder->BytesCount++;     // x is normal code symbol
                 return(ESC);         // = return x;
    case SDNORM: pRCoder->pacfeed = 1;     // x=End Of Code ESC sequence
                 pRCoder->BytesCount++;
                 return(0);
    case EOF:    ErrorMessage("EOF detected.\n");
                 exit(-1);
    default:     ErrorMessage("Unsupported ESC-code detected\n");
                 exit(-1);
  }
  return(0);
}

/* ---------------------------------------------------------- */

void RSendByte(RCENC *pRCoder, int x, int mode)
{
  pRCoder->BytesCount++;
  pRCoder->fByteOut(pRCoder->pCodeStream, x);
  
  if(x==ESC)
  {
    if (mode != 1)
    {
      pRCoder->BytesCount++;
      pRCoder->fByteOut(pRCoder->pCodeStream, STUFF);
    }
    else
    {
      pRCoder->BytesCount++;
      pRCoder->fByteOut(pRCoder->pCodeStream, SDNORM);
    }
  }
}

/* ---------------------------------------------------------- */

void RCEncode(RCENC *pRCoder, int pixel, QSMODEL *qsm)
{
  int syfreq, ltfreq;

  pixel &= pRCoder->mask;
  qsgetfreq(qsm,pixel,&syfreq,&ltfreq);
  encode_shift(pRCoder,syfreq,ltfreq);
  qsupdate(qsm,pixel);
}
/* ---------------------------------------------------------- */

int RCDecode(RCENC *pRCoder, QSMODEL *qsm)
{
  int  pixel,syfreq, ltfreq;

  ltfreq = decode_culshift(pRCoder, BITSINCNT);
  pixel = qsgetsym(qsm, ltfreq);
  qsgetfreq(qsm,pixel,&syfreq,&ltfreq);
  decode_update(pRCoder, syfreq, ltfreq, 1<<BITSINCNT);

  qsupdate(qsm,pixel);

  pixel &= pRCoder->mask;
  return pixel;
}

/* ---------------------------------------------------------- */
/* ---------------------------------------------------------- */

void InitEnc( RCENC *pRCoder, char c, int initlength)
{
  pRCoder->low = 0;                // Full code range
  pRCoder->range = Top_value;
  pRCoder->buffer = c;
  pRCoder->help = 0;               // No bytes to follow
  pRCoder->BytesCount = initlength;
}

/* ---------------------------------------------------------- */

static Inline void enc_normalize( RCENC *pRCoder )
{
  while(pRCoder->range <= Bottom_value)               // do we need renormalisation?
  {
    if (pRCoder->low < (CODE_VALUE)0xff<<SHIFT_BITS)  // no carry possible --> output
    {
      RSendByte(pRCoder, pRCoder->buffer, 0);
      for(; pRCoder->help; pRCoder->help--) RSendByte(pRCoder,0xff,0);    
      pRCoder->buffer = (unsigned char)(pRCoder->low >> SHIFT_BITS);
    }
    else 
      if (pRCoder->low & Top_value)                   // carry now, no future carry
      {
        RSendByte(pRCoder, pRCoder->buffer+1, 0);
        for(; pRCoder->help; pRCoder->help--) RSendByte(pRCoder, 0, 0);
        pRCoder->buffer = (unsigned char)(pRCoder->low >> SHIFT_BITS);
      }
      else                                        // passes on a potential carry
        pRCoder->help++;

    pRCoder->range <<= 8;
    pRCoder->low = (pRCoder->low<<8) & (Top_value-1);
  }
}

/* ---------------------------------------------------------- */

// Encode a symbol using frequencies
// pRCoder is the range coder to be used
// sy_f is the interval length (frequency of the symbol)
// lt_f is the lower end (frequency sum of < symbols)
// tot_f is the total interval length (total frequency sum)
// or (a lot faster): tot_f = 1<<shift
void encode_freq( RCENC *pRCoder, CODE_VALUE sy_f, CODE_VALUE lt_f, CODE_VALUE tot_f )
{
  CODE_VALUE r, tmp;
	
  enc_normalize( pRCoder );
	r = pRCoder->range / tot_f;
	tmp = r * lt_f;
	pRCoder->low += tmp;

  if (lt_f+sy_f < tot_f) pRCoder->range = r * sy_f;
  else pRCoder->range -= tmp;
}

/* ---------------------------------------------------------- */

void encode_shift( RCENC *pRCoder, CODE_VALUE sy_f, CODE_VALUE lt_f)
{
  CODE_VALUE r, tmp;

  enc_normalize( pRCoder );
	r = pRCoder->range>>BITSINCNT;
	tmp = r * lt_f;
	pRCoder->low += tmp;

  if ((lt_f+sy_f) >>BITSINCNT) pRCoder->range -= tmp;
	else pRCoder->range = r * sy_f;
}

/* ---------------------------------------------------------- */

ULONG RCFlushEncoder( RCENC *pRCoder )
{
  ULONG tmp;
  
  enc_normalize(pRCoder);     /* now we have a normalized state */
  if ((pRCoder->low & (Bottom_value-1)) < (pRCoder->BytesCount>>1)) tmp = pRCoder->low >> SHIFT_BITS;
  else tmp = (pRCoder->low >> SHIFT_BITS) + 1;
    
  if (tmp > 0xff) /* we have a carry */
  {
    RSendByte(pRCoder, pRCoder->buffer+1, 0);
    for(; pRCoder->help; pRCoder->help--) RSendByte(pRCoder, 0, 0);
  } 
  else  /* no carry */
  {
    RSendByte(pRCoder, pRCoder->buffer, 0);
    for(; pRCoder->help; pRCoder->help--) RSendByte(pRCoder, pRCoder->buffer, 0xff);
  }
  RSendByte(pRCoder, tmp&0xff, 0);
  RSendByte(pRCoder, ESC, 1);           // End-of-Code-Sequence

  tmp = (ULONG) pRCoder->BytesCount;
  pRCoder->BytesCount = 0;

  return(tmp);
}

/* ---------------------------------------------------------- */

int InitDec(RCENC *pRCoder)
{
  int c;
  
  pRCoder->BytesCount=0;
  c = RInputByte(pRCoder);
  
  if (c==EOF) return EOF;
  pRCoder->buffer = RInputByte(pRCoder);
  pRCoder->low = pRCoder->buffer >> (8-EXTRA_BITS);
  pRCoder->range = (CODE_VALUE)1 << EXTRA_BITS;
  pRCoder->pacfeed = 0;
  return c;
}

/* ---------------------------------------------------------- */

static Inline void dec_normalize( RCENC *pRCoder )
{
  while (pRCoder->range <= Bottom_value)
  {
    pRCoder->low = (pRCoder->low<<8) | ((pRCoder->buffer<<EXTRA_BITS)&0xff);
    pRCoder->buffer = RInputByte(pRCoder);
    pRCoder->low |= pRCoder->buffer >> (8-EXTRA_BITS);
    pRCoder->range <<= 8;
  }
}

/* ---------------------------------------------------------- */

// Calculate culmulative frequency for next symbol. Does NO update!
// pRCoder is the range coder to be used
// tot_f is the total frequency
// or: totf is 1<<shift
// returns the <= culmulative frequency
CODE_VALUE decode_culfreq( RCENC *pRCoder, CODE_VALUE tot_f )
{
  CODE_VALUE tmp;

  dec_normalize(pRCoder);
  pRCoder->help = pRCoder->range/tot_f;
  tmp = pRCoder->low/pRCoder->help;

  return (tmp>=tot_f ? tot_f-1 : tmp);
}

/* ---------------------------------------------------------- */

CODE_VALUE decode_culshift( RCENC *pRCoder, CODE_VALUE shift )
{
  CODE_VALUE tmp;

  dec_normalize(pRCoder);
  pRCoder->help = pRCoder->range>>shift;
  tmp = pRCoder->low/pRCoder->help;

  return (tmp>>shift ? ((CODE_VALUE)1<<shift)-1 : tmp);
}

/* ---------------------------------------------------------- */

// Update decoding state
// pRCoder is the range coder to be used
// sy_f is the interval length (frequency of the symbol)
// lt_f is the lower end (frequency sum of < symbols)
// tot_f is the total interval length (total frequency sum)
void decode_update( RCENC *pRCoder, CODE_VALUE sy_f, CODE_VALUE lt_f, CODE_VALUE tot_f)
{   
  CODE_VALUE tmp;
  
  tmp = pRCoder->help * lt_f;
  pRCoder->low -= tmp;

  if (lt_f + sy_f < tot_f) pRCoder->range = pRCoder->help * sy_f;
  else pRCoder->range -= tmp;
}

/* ---------------------------------------------------------- */

// Decode a byte/short without modelling
// pRCoder is the range coder to be used
unsigned char decode_byte(RCENC *pRCoder)
{ 
  unsigned char tmp = decode_culshift(pRCoder,8);
  decode_update( pRCoder,1,tmp,(CODE_VALUE)1<<8);
  return tmp;
}

/* ---------------------------------------------------------- */

unsigned short decode_short(RCENC *pRCoder)
{ 
  unsigned short tmp = decode_culshift(pRCoder,16);
  decode_update(pRCoder,1,tmp,(CODE_VALUE)1<<16);
  return tmp;
}

/* ---------------------------------------------------------- */

ULONG RCFlushDecoder(RCENC *pRCoder )
{
  ULONG tmp;

  dec_normalize(pRCoder);             // normalize to use up all bytes
  tmp = pRCoder->BytesCount;
  pRCoder->BytesCount = 0;
  return tmp;
}

/* ---------------------------------------------------------- */

// retrieval of estimated frequencies for a symbol
// m   qsmodel to be questioned
// sym  symbol for which data is desired; must be <n
// sy_f frequency of that symbol
// lt_f frequency of all smaller symbols together
// the total frequency is 1<<lg_totf
void qsgetfreq(QSMODEL *m, int sym, int *sy_f, int *lt_f )
{
  *sy_f = m->cf[sym+1] - (*lt_f = m->cf[sym]);
}	

/* ---------------------------------------------------------- */

// find out symbol for a given cumulative frequency
// m   qsmodel to be questioned
// lt_f  cumulative frequency
int qsgetsym(QSMODEL *m, int lt_f )
{
  int lo, mid, hi;
  UINT16 *tmp;
  
  tmp = m->search+(lt_f>>m->searchshift);
  lo = *tmp;
  hi = *(tmp+1) + 1;
  while (lo+1 < hi )
  {
    mid = (lo+hi)>>1;
    if (lt_f < m->cf[mid]) hi = mid;
    else lo = mid;
  }
  return lo;
}

/* ---------------------------------------------------------- */

// update model
// m   qsmodel to be updated
// sym  symbol that occurred (must be <n from init)
void qsupdate(QSMODEL *m, int sym )
{   
  if (m->left <= 0) dorescale(m);
  m->left--;
  m->newf[sym] += m->incr;
}


/*=========================== User part =============================*/

void RCInitEncoder(RCENC *pRCEnc, FLOW *f, int Depth)
{
  pRCEnc->mask     = (1<<Depth)-1;
  pRCEnc->fByteOut = f->OutputByte;
  pRCEnc->pCodeStream = f;
  InitEnc(pRCEnc,0,0);
}

void RCInitDecoder(RCENC *pRCDec, FLOW *f, int Depth)
{
  pRCDec->mask     = (1<<Depth)-1;
  pRCDec->fByteIn  = f->InputByte;
  pRCDec->pCodeStream = f;
  InitDec(pRCDec);
}
