/*-------------------------------------------------------------------*/
/* COLOR.C     Pavel Kopylov                                         */
/*                                                                   */
/*-------------------------------------------------------------------*/

#define ProgName        "COLOR"
#define VersionNumber   "0.03"
#define LastUpdated     "11.03.2004"

/* ------------------------------------------------------------------*/

#include <math.h>
#include "owntypes.h"
#include "interfc.h"
#include "color.h"

double WhitePointData[10][3]= 
{
  {109.850, 100.00, 35.585}, // the CIE Illuminant A, light from a tungsten lamp (Incandescent), at 500 watts
  {98.074, 100.00, 118.232}, // the CIE illuminant C, average sunlight, at 500 watts;
  {96.422, 100.00, 82.521},  // 'D50' or 'D5000', the CIE illuminant used in graphics printing,
  {95.682, 100.00, 92.149},  // 'D55' CIE illuminant that approximates a cloudy bright day;
  {95.047, 100.00, 108.883}, // 'D65' or 'D6500', the CIE illuminant that approximates daylight;
  {100.0, 100.0, 100.0},     //  the CIE illuminant E, normalized reference source.
  {94.972, 100.0, 122.638},  //  'D75'
  {99.187, 100.0, 67.395},   //  'F2' Fluoriscent
  {95.044, 100.0, 108.755},  //  'F7'
  {100.966, 100.0, 64.370},  //  'F10'
};                                    

double r_cubert(double x);
void xyz_to_uvwprime(double xcap, double ycap, double zcap, double *uprime, double *vprime, double *wprime);
void uvprimey_to_xyz(double uprime, double vprime, double *xcap, double *ycap, double *zcap);


/* ------------------------------------------------------------------*/
void RGB_to_XYZ(BYTE inp_R, BYTE inp_G, BYTE inp_B, double *xcap, double *ycap, double *zcap)
{
  double var_R, var_G, var_B;


  var_R = ( (double)inp_R/255.0 );        //inp_R = From 0 to 255
  var_G = ( (double)inp_G/255.0 );        //inp_G = From 0 to 255
  var_B = ( (double)inp_B/255.0 );        //inp_B = From 0 to 255

  if (var_R > 0.03928) var_R = pow((var_R+0.055)/1.055,2.4);
  else                 var_R = var_R/12.92;
  
  if (var_G > 0.03928) var_G = pow((var_G+0.055)/1.055,2.4);
  else                 var_G = var_G/12.92;
  
  if (var_B > 0.03928) var_B = pow((var_B+0.055)/1.055,2.4);
  else                 var_B = var_B/12.92;

  var_R = var_R * 100;
  var_G = var_G * 100;
  var_B = var_B * 100;

  //Observer. = 2, Illuminant = D65
  (*xcap) = var_R * 0.412453 + var_G * 0.35758  + var_B * 0.180423;
  (*ycap) = var_R * 0.212671 + var_G * 0.71516  + var_B * 0.072169;
  (*zcap) = var_R * 0.019334 + var_G * 0.119193 + var_B * 0.950227;
}

/* ------------------------------------------------------------------*/
void XYZ_to_LAB(WHITEPOINT wp, double xcap, double ycap, double zcap, double *lstar, double *astar, double *bstar)
{
  double fy;
  double xxn, yyn, zzn;


  xxn=xcap/WhitePointData[wp][0];
  yyn=ycap/WhitePointData[wp][1];
  zzn=zcap/WhitePointData[wp][2];
/*
  if(xxn>0.008856) xxn=r_cubert(xxn);
  else             xxn=(7.787*xxn)+(16.0/116.0);

  if(yyn>0.008856) yyn=r_cubert(yyn);
  else             yyn=(7.787*yyn)+(16.0/116.0);

  if(zzn>0.008856) zzn=r_cubert(zzn);
  else             zzn=(7.787*zzn)+(16.0/116.0);

  (*lstar) = (116*yyn)-16;
  (*astar) = 500*(xxn-yyn);
  (*bstar) = 200*(yyn-zzn);
*/

  if (yyn>216.0/24389.0) (*lstar) = 116*r_cubert(yyn)-16.0;      //216/24389 = 0.008856
  else (*lstar) = yyn*24389.0/27.0;

  fy = ((*lstar)+16.0)/116.0;

  if (xxn>216.0/24389.0) (*astar) = 500*(r_cubert(xxn)-fy);
  else (*astar) = 500*(841.0/108.0*xxn - fy + 4.0/29.0);

  if (zzn>216.0/24389.0) (*bstar) = 200*(fy - r_cubert(zzn));
  else (*bstar) = 200*(fy - 4.0/29.0 - 841.0/108.0*zzn);
}

/* ------------------------------------------------------------------*/
void XYZ_to_LUV(WHITEPOINT wp, double xcap, double ycap, double zcap, double *lstar, double *ustar, double *vstar)
{
  double uprime, unprime, vprime, vnprime, wprime, wnprime;
  double x;


  x=ycap/WhitePointData[wp][1];

  if (x>216.0/24389.0) (*lstar) = 116*r_cubert(x)-16;      //216/24389 = 0.008856
  else (*lstar) = x*24389.0/27.0;

  // Compute (un',vn') from Whitepoint.
  xyz_to_uvwprime(WhitePointData[wp][0], WhitePointData[wp][1], WhitePointData[wp][2], &unprime, &vnprime, &wnprime);

  // Compute (u',v') from (XCAP,YCAP,ZCAP).
  xyz_to_uvwprime(xcap, ycap, zcap, &uprime, &vprime, &wprime);
  
  // Compute (u*,v*) from (u',v') and (un',vn').
  (*ustar) = 13.0 * (*lstar) * (uprime-unprime);
  (*vstar) = 13.0 * (*lstar) * (vprime-vnprime);
}

/* ------------------------------------------------------------------*/
void LAB_to_XYZ(WHITEPOINT wp, double lstar, double astar, double bstar, double *xcap, double *ycap, double *zcap)
{

  double lf;


  lf = (lstar+16)/116.0;

  if (lstar>=29*bstar/50+8) (*zcap) = pow(lf-bstar/200,3)*WhitePointData[wp][2];
  else (*zcap) = (108.0/841.0)*(lf - 4.0/29.0 - bstar/200.0)*WhitePointData[wp][2];

  if (lstar>=8-29*astar/125.0) (*xcap) = pow(astar/500.0+lf,3.0)*WhitePointData[wp][0];
  else (*xcap) = (108.0/841.0)*(astar/500.0 + lf - 4.0/29.0)*WhitePointData[wp][0];

  if (lstar>=8) (*ycap) = pow(lf,3)*WhitePointData[wp][1];
  else (*ycap) = 27*lstar/24389.0*WhitePointData[wp][1];
}


/* ------------------------------------------------------------------*/
void LUV_to_XYZ(WHITEPOINT wp, double lstar, double ustar, double vstar, double *xcap, double *ycap, double *zcap)
{

  double uprime, unprime, vprime, vnprime, wnprime;
  double lf;


  lf = (lstar+16)/116.0;
  
  if (lstar>=8) (*ycap) = pow(lf,3)*WhitePointData[wp][1];
  else (*ycap) = 27*lstar/24389.0*WhitePointData[wp][1];
 
  // Compute(un',vn')from(XCAPN,YCAPN,ZCAPN).
  xyz_to_uvwprime(WhitePointData[wp][0], WhitePointData[wp][1], WhitePointData[wp][2], &unprime, &vnprime, &wnprime);

  // Compute(u',v')from(un',vn')and(l*,u*,v*).
  uprime = unprime + ustar /(13.0 * lstar); 
  vprime = vnprime + vstar /(13.0 * lstar);

  // Now compute XCAP and ZCAP from(u',v')and YCAP.
  uvprimey_to_xyz(uprime, vprime, xcap, ycap, zcap);
}

/* ------------------------------------------------------------------*/
void XYZ_to_RGB(double xcap, double ycap, double zcap, BYTE *r, BYTE *g, BYTE *b)
{
  double var_X, var_Y, var_Z;
  double var_R, var_G, var_B;


  var_X = xcap / 100.0;        //inp_X = From 0 to ref_X
  var_Y = ycap / 100.0;        //inp_Y = From 0 to ref_Y
  var_Z = zcap / 100.0;        //inp_Z = From 0 to ref_Y

  var_R = var_X *  3.2410 + var_Y * -1.5374 + var_Z * -0.4986;
  if (var_R<0) var_R=0;
  var_G = var_X * -0.9692 + var_Y *  1.8760 + var_Z *  0.0416;
  if (var_G<0) var_G=0;
  var_B = var_X *  0.0556 + var_Y * -0.2040 + var_Z *  1.0570;
  if (var_B<0) var_B=0;

  if ( var_R > 0.00304 ) var_R = 1.055*pow(var_R,(1/2.4)) - 0.055;
  else                   var_R = 12.92*var_R;
  if ( var_G > 0.00304 ) var_G = 1.055*pow(var_G,(1/2.4)) - 0.055;
  else                   var_G = 12.92*var_G;
  if ( var_B > 0.00304 ) var_B = 1.055*pow(var_B,(1/2.4)) - 0.055;
  else                   var_B = 12.92*var_B;

  if (var_R<0) var_R = - var_R;
  if (var_G<0) var_G = - var_G;
  if (var_B<0) var_B = - var_B;

  if (var_R>1) var_R=1;
  if (var_G>1) var_G=1;
  if (var_B>1) var_B=1;

  (*r) = (BYTE)(var_R * 255);
  (*g) = (BYTE)(var_G * 255);
  (*b) = (BYTE)(var_B * 255);
}

void GetWhitepoint(WHITEPOINT wp, double *xcapn, double *ycapn, double *zcapn)
{
  (*xcapn) = WhitePointData[wp][0];
  (*ycapn) = WhitePointData[wp][1];
  (*zcapn) = WhitePointData[wp][2];
}



double ColorDistance76(WHITEPOINT wp, BYTE r1, BYTE g1, BYTE b1, BYTE r2, BYTE g2, BYTE b2)
{
  double x1,y1,z1,x2,y2,z2;
  double l1,a1,b1b,l2,a2,b2b;
  double de;

  if(r1==0 && g1==0 && b1==0){r1=1; g1=1; b1=1;}
  if(r2==0 && g2==0 && b2==0){r2=1; g2=1; b2=1;}

  RGB_to_XYZ(r1, g1, b1, &x1, &y1, &z1);
  RGB_to_XYZ(r2, g2, b2, &x2, &y2, &z2);

  XYZ_to_LAB(wp, x1, y1, z1, &l1, &a1, &b1b);
  XYZ_to_LAB(wp, x2, y2, z2, &l2, &a2, &b2b);

  de = DeltaE76(l1, a1, b1b, l2, a2, b2b);

  return de;
}

double ColorDistance94(WHITEPOINT wp, BYTE r1, BYTE g1, BYTE b1, BYTE r2, BYTE g2, BYTE b2)
{
  double x1,y1,z1,x2,y2,z2;
  double l1,a1,b1b,l2,a2,b2b;
  double de;


  RGB_to_XYZ(r1, g1, b1, &x1, &y1, &z1);
  RGB_to_XYZ(r2, g2, b2, &x2, &y2, &z2);

  XYZ_to_LAB(wp, x1, y1, z1, &l1, &a1, &b1b);
  XYZ_to_LAB(wp, x2, y2, z2, &l2, &a2, &b2b);

  de = DeltaE94(l1, a1, b1b, l2, a2, b2b);

  return de;
}


double DeltaE76(double l1, double a1, double b1, double l2, double a2, double b2)
{
  return (sqrt((l1-l2)*(l1-l2)+(a1-a2)*(a1-a2)+(b1-b2)*(b1-b2)));
}

double DeltaE94(double l1, double a1, double b1, double l2, double a2, double b2)
{
  double dL, dA, dB, C1, C2, dC, dH, Sl, Sc, Sh;

  dL = l1-l2;
  dA = a1-a2;
  dB = b1-b2;

  C1 = sqrt(a1*a1+b1*b1);
  C2 = sqrt(a2*a2+b2*b2);
  dC = C1-C2;

  dH = sqrt(dA*dA+dB*dB+dC*dC);

  Sl = dL/2;
  Sc = dC/(1+0.048*C1);
  Sh = dH/(1+0.014*C1);

  return sqrt(Sl*Sl+Sc*Sc+Sh*Sh);
}


/* ------------------------------------------------------------------*/
/* ------------------------------------------------------------------*/
/* ------------------------------------------------------------------*/
double r_cubert(double x)
{
  if(x > 0.0)return(pow(x,(1.0/3.0)));
  else 
    if(x == 0.0)return(0.0);
  else
    return(-pow(fabs(x),(1.0/3.0)));
}

void xyz_to_uvwprime(double xcap, double ycap, double zcap, double *uprime, double *vprime, double *wprime)
{
  double denom;
  
  denom = xcap + 15.0 * ycap + 3.0 * zcap;

  if(denom == 0.0)
  {
    (*uprime) = 4.0;
    (*vprime) = 9.0;
    (*wprime) = 3.0;
  }
  else
  {
    (*uprime) = 4.0 * xcap / denom;
    (*vprime) = 9.0 * ycap / denom;
    (*wprime) =(- 3.0 * xcap + 6.0 * ycap + 3.0 * zcap)/ denom;
  }
}

void uvprimey_to_xyz(double uprime, double vprime, double *xcap, double *ycap, double *zcap)
{
 (*xcap)= 9.0 * (*ycap) * uprime /(4.0 * vprime);
 (*zcap)= -(*xcap)/ 3.0 - 5.0 * (*ycap) + 3.0 * (*ycap) / vprime;
}
