/****************************************************************

    DISTANCE.C

    Ismo Krkkinen

    11.6.2001

    Provides generalized distance function that allows also user-
    defined distances via function. Note that it is not assumed
    that functions are metrics, therefore the direction needs to
    be specified.

****************************************************************/

#include "distance.h"
#include <math.h>
#include "owntypes.h"
#include "interfc.h"
#include "stcomplx.h"
#include "mdl.h"
#include "memctrl.h"

static double EuclideanSquareD2D(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(TS, To);
    Direction = VectorSize(TS);
    while (Direction--) {
	Diff = (double)(*F++ - *T++);
	Sum += Diff * Diff;
    }
    return Sum;
}

static double EuclideanSquareD2M(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(CB, To);
    Direction = VectorSize(TS);
    while (Direction--) {
	Diff = (double)(*F++ - *T++);
	Sum += Diff * Diff;
    }
    return Sum;
}

static double EuclideanSquareM2D(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(TS, To);
    Direction = VectorSize(TS);
    while (Direction--) {
	Diff = (double)(*F++ - *T++);
	Sum += Diff * Diff;
    }
    return Sum;
}

static double EuclideanSquareM2M(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(CB, To);
    Direction = VectorSize(CB);
    while (Direction--) {
	Diff = (double)(*F++ - *T++);
	Sum += Diff * Diff;
    }
    return Sum;
}

/* Specialized versions for some fixed dimensionalities. */
/* Below are "templates", copy and replace XYZZY with dimensionality. */

/*
static double EuclideanSquareD2D_XYZZY(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(TS, To);
    Direction = XYZZY;
    while (Direction--) {
        Diff = (double)(*F++ - *T++);
        Sum += Diff * Diff;
    }
    return Sum;
}


static double EuclideanSquareD2M_XYZZY(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(CB, To);
    Direction = XYZZY;
    while (Direction--) {
        Diff = (double)(*F++ - *T++);
        Sum += Diff * Diff;
    }
    return Sum;
}

static double EuclideanSquareM2D_XYZZY(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(TS, To);
    Direction = XYZZY;
    while (Direction--) {
        Diff = (double)(*F++ - *T++);
        Sum += Diff * Diff;
    }
    return Sum;
}

static double EuclideanSquareM2M_XYZZY(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(CB, To);
    Direction = XYZZY;
    while (Direction--) {
        Diff = (double)(*F++ - *T++);
        Sum += Diff * Diff;
    }
    return Sum;
}

End of templates for Euclidean square functions. */

static double EuclideanSquareD2D_2(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(TS, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}


static double EuclideanSquareD2M_2(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(CB, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}

static double EuclideanSquareM2D_2(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(TS, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}

static double EuclideanSquareM2M_2(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(CB, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}


static double EuclideanSquareD2D_3(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(TS, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}


static double EuclideanSquareD2M_3(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(CB, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}

static double EuclideanSquareM2D_3(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(TS, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}

static double EuclideanSquareM2M_3(TRAININGSET* TS, CODEBOOK* CB,
    PARTITIONING* PA, int From, int To, int Direction)
{
    double Sum = 0;
    double Diff;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(CB, To);
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    Diff = (double)(*F++ - *T++);
    Sum += Diff * Diff;
    return Sum;
}


static double EuclideanD2D(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    return (double)sqrt(EuclideanSquareD2D(TS, CB, PA, From, To, Direction));
}

static double EuclideanD2M(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    return (double)sqrt(EuclideanSquareD2M(TS, CB, PA, From, To, Direction));
}

static double EuclideanM2D(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    return (double)sqrt(EuclideanSquareM2D(TS, CB, PA, From, To, Direction));
}

static double EuclideanM2M(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    return (double)sqrt(EuclideanSquareM2M(TS, CB, PA, From, To, Direction));
}

static double ManhattanD2D(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(TS, To);
    Direction = VectorSize(TS);
    while (Direction--) Sum += abs(*F++ - *T++);
    return Sum;
}

static double ManhattanD2M(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(TS, From);
    T = Vector(CB, To);
    Direction = VectorSize(TS);
    while (Direction--) Sum += abs(*F++ - *T++);
    return Sum;
}

static double ManhattanM2D(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(TS, To);
    Direction = VectorSize(TS);
    while (Direction--) Sum += abs(*F++ - *T++);
    return Sum;
}

static double ManhattanM2M(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double Sum = 0;
    VECTORTYPE F;
    VECTORTYPE T;
    F = Vector(CB, From);
    T = Vector(CB, To);
    Direction = VectorSize(CB);
    while (Direction--) Sum += abs(*F++ - *T++);
    return Sum;
}

#ifdef LOG2
#undef LOG2
#endif
#define LOG2(a) ((a) == 0 ? -1e10 : 1.4426950408889634073599246810019 * log(a))

/* Presumes that all values are either 0 or TS->MaxValue. */
static double Codelength(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    double CL = 0;
    double POne, TotalInv;
    int k;
    VECTORTYPE x;
    x = Vector(TS, From);
    TotalInv = (double)(1.0 / TS->MaxValue);
    for (k = 0; k < VectorSize(TS); k++) {
	POne = VectorScalar(CB, To, k) * TotalInv;
	if (*x++) CL += (double)LOG2(POne);
	else CL += (double)LOG2(1.0 - POne);
    }
    return (double)(-CL - LOG2(CCFreq(PA, To) / (double) TotalFreq(TS)));
}

#undef LOG2


DistanceInfo* diNew(TRAININGSET* TS, DT_TYPE Type,
    DT_DistanceFunction Data2Data, DT_DistanceFunction Data2Model,
    DT_DistanceFunction Model2Data, DT_DistanceFunction Model2Model)
{
    DistanceInfo* di = allocate(sizeof(DistanceInfo));
    switch (Type) {
    case DT_User:
	if (!Data2Data || !Data2Model || !Model2Data || !Model2Model) {
	    ErrorMessage("distance.c::diNew: no user functions given.\n");
	    ExitProcessing(FATAL_ERROR);
	}
	break;
    case DT_Euclidean:
	Data2Data = &EuclideanD2D;
	Data2Model = &EuclideanD2M;
	Model2Data = &EuclideanM2D;
	Model2Model = &EuclideanM2M;
	break;
    case DT_EuclideanSquare:
	switch(VectorSize(TS)) {
	case 2:
	    Data2Data = &EuclideanSquareD2D_2;
	    Data2Model = &EuclideanSquareD2M_2;
	    Model2Data = &EuclideanSquareM2D_2;
	    Model2Model = &EuclideanSquareM2M_2;
	    break;
	case 3:
	    Data2Data = &EuclideanSquareD2D_3;
	    Data2Model = &EuclideanSquareD2M_3;
	    Model2Data = &EuclideanSquareM2D_3;
	    Model2Model = &EuclideanSquareM2M_3;
	    break;
	default:
	    Data2Data = &EuclideanSquareD2D;
	    Data2Model = &EuclideanSquareD2M;
	    Model2Data = &EuclideanSquareM2D;
	    Model2Model = &EuclideanSquareM2M;
	    break;
	}
	break;
    case DT_Manhattan:
	Data2Data = &ManhattanD2D;
	Data2Model = &ManhattanD2M;
	Model2Data = &ManhattanM2D;
	Model2Model = &ManhattanM2M;
	break;
    case DT_DeltaSC:
	Data2Data = &StochasticComplexityDelta;
	Data2Model = &StochasticComplexityDelta;
	Model2Data = &StochasticComplexityDelta;
	Model2Model = &StochasticComplexityDelta;
	break;
    case DT_CodeLength:
	Data2Data = &Codelength;
	Data2Model = &Codelength;
	Model2Data = &Codelength;
	Model2Model = &Codelength;
	break;
    case DT_DeltaMDL:
	Data2Data = &MDLDelta;
	Data2Model = &MDLDelta;
	Model2Data = &MDLDelta;
	Model2Model = &MDLDelta;
	break;
    default:
	ErrorMessage("distance.c::diNew: unknown distance function.\n");
	ExitProcessing(FATAL_ERROR);
	break;
    }
    di->Type = Type;
    di->F[DT_DATA2DATA] = Data2Data;
    di->F[DT_DATA2MODEL] = Data2Model;
    di->F[DT_MODEL2DATA] = Model2Data;
    di->F[DT_MODEL2MODEL] = Model2Model;
    return di;
}

void diDelete(DistanceInfo* DI) {
    deallocate(DI);
}

/*
double diDistance(DistanceInfo* DI,
    TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA,
    int From, int To, int Direction)
{
    return (DI->Func)(TS, CB, PA, From, To, Direction);
}
*/

