/*-------------------------------------------------------------------*/
/* IMAGE.C     Pavel Kopylov                                         */
/*                                                                   */
/* Originated from "binimage" by E.Ageenko                           */
/*-------------------------------------------------------------------*/

#define ProgName        "IMAGE"
#define VersionNumber   "0.06"
#define LastUpdated     "31.05.2003"

/* ------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "file.h"
#include "image.h"
#include "pgm.h"
#include "memctrl.h"
#include "interfc.h"
#include "owntypes.h"


/* ==================== ERROR HANDLING ==============================*/

char* E_NOMEMORY     = "Not enough memory to complete operation, aborted.";
char* E_IMSIZE       = "Invalid image size.";
char* E_IMBOUNDS     = "Bounds can not exseed image size.";
char* E_UBIMAGEMODEL = "Unknow image representation model.";
char* E_LINEPASSED   = "Line is already left buffer and can't be observed.";
char* E_LINESTORED   = "Line is already stored and can't be observed.";
char* E_UFMODE       = "Unknow file open mode.";
char* E_UBUFFERING   = "Unknow image buffering method.";
char* E_LARGEBUFFER  = "Module supports buffer size up to XXXX lines only.";
char* E_DEFAULT      = "Internal error";


/*----------------- Internal definitiones  -----------------------------------*/

void ReadLines(IMAGE *image, ULONG n);
void WriteLines(IMAGE *image, ULONG n);
void EmptyLine(IMAGE *image, ULONG line);

/* ====================== internal ==================================*/
ULONG BytesPerLine(IMAGE* image)
{
  ULONG    NBytes;

  switch(image->ModeType)
  {
    case BINMODE:
      NBytes = 1 + (ULONG)((image->ImageSizeX-1)/8);
      break;
    case GRAYMODE:
      NBytes = (ULONG)image->ImageSizeX;
      break;
    case RGBMODE:
      NBytes = (ULONG)(image->ImageSizeX*RGBSIZE);     // = ULONG, usually = 4 bytes
      break;
  }
  return (NBytes);
}


/*===================================================================*/
/* ==================  Internal I/O routines  =======================*/
/*===================================================================*/
void EmptyLine(IMAGE *image, ULONG line)
{
  ULONG  x;
  BYTE   filler, r, g, b;


  switch(image->ModeType)
  {
    case BINMODE:
      filler = ((BYTE)image->DefaultColor == BIN_WHITE) ? 0 : 0xFF;
      break;
    case GRAYMODE:
      filler = (BYTE)(image->DefaultColor & 0xFF);
      break;
    case RGBMODE:
      r = (BYTE)RED(image->DefaultColor);
      g = (BYTE)GREEN(image->DefaultColor);
      b = (BYTE)BLUE(image->DefaultColor);
      for(x=0; x<image->ImageBytesX; x+=RGBSIZE)
      {
        image->data[line][x]  = r;
        image->data[line][x+1]= g;
        image->data[line][x+2]= b;
      }
      break;
  }

  if(image->ModeType != RGBMODE)
    for(x=0; x<image->ImageBytesX; x++)
    {
      image->data[line][x]=filler;
    }
}


/* ------------------------------------------------------------------*/
YESNO EmptyWholeImage(IMAGE* image)
{
  ULONG y;

  if(image->Buffering == WHOLEIMAGE)
  {
    for(y=0; y<image->LinesInBuffer; y++)
    {
      EmptyLine(image,y);
    }
    return (YES);
  }
  else return (NO);
}


/* ------------------------------------------------------------------*/
void ReadLines(IMAGE* image, ULONG n)
{
  long      tmp;
  ULONG     k,i;
  BYTE      *RData, *GData, *BData;


  if(image->ModeType == RGBMODE)
  {
    RData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
    GData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
    BData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
  }

  for(k=0; k<n; k++)
  {
    switch(image->ModeType)
    {
      case BINMODE:
        ReadPbmBitLine(image->FilePointer, image->data[image->BufCount], image->ImageBytesX); 
        break;
      case GRAYMODE: 
        ReadPgmLine(image->FilePointer, image->data[image->BufCount], image->ImageBytesX);
        break;
      case RGBMODE: 
        ReadPpmLine(image->FilePointer, RData, GData, BData, image->ImageSizeX);
        for(i=0;i<image->ImageSizeX;i++)
        {
          image->data[image->BufCount][i*RGBSIZE]  = RData[i];
          image->data[image->BufCount][i*RGBSIZE+1]= GData[i];
          image->data[image->BufCount][i*RGBSIZE+2]= BData[i];
        }
        break;
     }
     image->BufCount++;
     if((tmp=image->BufCount-image->LinesInBuffer) >= 0) image->BufCount = tmp;
  }

  if(image->ModeType == RGBMODE)
  {
    deallocate(RData);
    deallocate(GData);
    deallocate(BData);
  }

  image->FirstBufLine += n;
  image->LastBufLine  += n;
  image->LinesReadTotal += n;
}


/* ------------------------------------------------------------------*/
void WriteLines(IMAGE* image, ULONG n)
{
  ULONG     k,i;
  long      tmp;
  BYTE      *RData, *GData, *BData;


  if(image->ModeType == RGBMODE)
  {
    RData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
    GData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
    BData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
  }

  for (k=0; k<n; k++)
  {
    switch (image->ModeType)
    {
      case BINMODE: 
        WritePbmBitLine(image->FilePointer, image->data[image->BufCount], image->ImageBytesX); 
        break;
      case GRAYMODE: 
        WritePgmLine(image->FilePointer, image->data[image->BufCount], image->ImageBytesX); 
        break;
      case RGBMODE: 
        for(i=0;i<image->ImageSizeX;i++)
        {
          RData[i] = image->data[image->BufCount][i*RGBSIZE];
          GData[i] = image->data[image->BufCount][i*RGBSIZE+1];
          BData[i] = image->data[image->BufCount][i*RGBSIZE+2];
        }
        WritePpmLine(image->FilePointer, RData, GData, BData, image->ImageSizeX); 
        break;
     }
     EmptyLine(image, image->BufCount);
     image->BufCount++;
     if((tmp=image->BufCount-image->LinesInBuffer) >= 0) image->BufCount = tmp;
  }

  if(image->ModeType == RGBMODE)
  {
    deallocate(RData);
    deallocate(GData);
    deallocate(BData);
  }

  image->FirstBufLine += n;
  image->LastBufLine  += n;
  image->LinesReadTotal += n;
}


/*===================================================================*/
/*==================== Image-Interface ==============================*/
/*===================================================================*/
int GetBitPixel(IMAGE* image, ULONG x, ULONG y)
{
  long by,bx,tmp;
  int value;

  if(image->ModeType!=BINMODE) return(-1);

  if(x<image->Bound.Left || x>image->Bound.Right || y<image->Bound.Upper || y>image->Bound.Lower)
    return(0);

  if(y < image->FirstBufLine)
  {
    ErrorMessage(E_LINEPASSED);
    exit(-1);
  }

  if((tmp=y-image->LastBufLine) > 0)
  {
    tmp += min(image->ScrollSize,image->ImageSizeY-y);
    if(image->FileMode==OUTPUT) WriteLines (image, tmp);
    else ReadLines (image, tmp);
  }

  by = (y - image->FirstBufLine + image->BufCount);
  if((tmp=by-image->LinesInBuffer) >= 0) by=tmp;

  bx= (x-1)>>3;
  value= image->data[by][bx];
  value >>= (7-((x-1)&7));
 
  return( (value & 1) != 0 ? BIN_BLACK : BIN_WHITE);
}


/*-------------------------------------------------------------------*/
int GetGrayPixel(IMAGE* image, ULONG x, ULONG y)
{
  long by,tmp;

  if(image->ModeType!=GRAYMODE) return(-1);
  if(x<image->Bound.Left || x>image->Bound.Right || y<image->Bound.Upper || y>image->Bound.Lower)
    return(0);

  if(y < image->FirstBufLine)
  {
    ErrorMessage(E_LINEPASSED);
    exit(-1);
  }

  if( (tmp=y-image->LastBufLine) > 0)
  {
    tmp += min(image->ScrollSize,image->ImageSizeY-y);
    if(image->FileMode==OUTPUT) WriteLines (image, tmp);
    else ReadLines (image, tmp);
  }

  by = (y - image->FirstBufLine + image->BufCount);
  if((tmp=by-image->LinesInBuffer) >= 0) by=tmp;

  return( (int) (image->data[by][x-1] ) );
}


/*-------------------------------------------------------------------*/
RGBPIXEL GetRGBPixel(IMAGE* image, ULONG x, ULONG y)
{
  long      by,tmp;
  BYTE      r,g,b;
  RGBPIXEL  tmpPixel;

  if(image->ModeType!=RGBMODE) return(RGB_BLACK);
  if(x<image->Bound.Left || x>image->Bound.Right || y<image->Bound.Upper || y>image->Bound.Lower) 
    return(RGB_BLACK);

  if(y < image->FirstBufLine)
  {
    ErrorMessage(E_LINEPASSED);
    exit(-1);
  }

  if( (tmp=y-image->LastBufLine) > 0)
  {
    tmp += min(image->ScrollSize,image->ImageSizeY-y);
    if(image->FileMode==OUTPUT) WriteLines(image, tmp);
    else ReadLines(image, tmp);
  }

  by= (y - image->FirstBufLine + image->BufCount);
  if( (tmp=by-image->LinesInBuffer) >= 0) by=tmp;

  r = image->data[by][(x-1)*RGBSIZE];
  g = image->data[by][(x-1)*RGBSIZE+1];
  b = image->data[by][(x-1)*RGBSIZE+2];
  tmpPixel = RGB(r,g,b);

  return (tmpPixel);
}

/*===================================================================*/
/*-------------------- Put Pixel ------------------------------------*/
/*===================================================================*/
void PutBitPixel(IMAGE* image, ULONG x, ULONG y, int pixel)
{
  long by,bx,tmp;
  BYTE mask;

  if(image->ModeType!=BINMODE) return;

  if(x<image->Bound.Left || x>image->Bound.Right || y<image->Bound.Upper || y>image->Bound.Lower) 
    return;

  if(y < image->FirstBufLine)
  {
    ErrorMessage(E_LINESTORED);
    exit(-1);
  }

  if((tmp = y-image->LastBufLine) > 0)
  {
    if(image->FilePointer)
		WriteLines (image, tmp);
    tmp += min(image->ScrollSize,image->ImageSizeY-y);
  }

  by = (y - image->FirstBufLine + image->BufCount);
  if((tmp = by-image->LinesInBuffer) >= 0) by=tmp;

  bx= (x-1)>>3;
  mask=0; mask++; 
  mask<<= (7-((x-1)&7));

  if(pixel!=BIN_WHITE) image->data[by][bx] |= mask;    /* set bit */
  else                 image->data[by][bx] &= ~mask;   /* kill bit */
}


/*-------------------------------------------------------------------*/
void PutGrayPixel(IMAGE* image, ULONG x, ULONG y, int pixel)
{
  long by,tmp;

  if(image->ModeType!=GRAYMODE) return;
  if(x<image->Bound.Left || x>image->Bound.Right || y<image->Bound.Upper || y>image->Bound.Lower)
    return;

  if(y < image->FirstBufLine)
  {
    ErrorMessage(E_LINESTORED);
    exit(-1);
  }

  if( (tmp=y-image->LastBufLine) > 0)
  {
    tmp += min(image->ScrollSize,image->ImageSizeY-y);
    WriteLines (image, tmp);
  }

  by= (y - image->FirstBufLine + image->BufCount);
  if( (tmp=by-image->LinesInBuffer) >= 0) by=tmp;

  image->data[by][x-1]= (BYTE)pixel;
}


/*-------------------------------------------------------------------*/
void PutRGBPixel(IMAGE* image, ULONG x, ULONG y, RGBPIXEL pixel)
{
  long  by,tmp;
  BYTE   r,g,b;

  if(image->ModeType!=RGBMODE) return;
  if(x<image->Bound.Left || x>image->Bound.Right || y<image->Bound.Upper || y>image->Bound.Lower)
   return;

  if(y < image->FirstBufLine)
  {
    ErrorMessage(E_LINESTORED);
    exit(-1);
  }

  if((tmp=y-image->LastBufLine) > 0)
  {
    tmp += min(image->ScrollSize,image->ImageSizeY-y);
    WriteLines (image, tmp);
  }

  by = (y - image->FirstBufLine + image->BufCount);
  if((tmp=by-image->LinesInBuffer) >= 0) by=tmp;

  r = (BYTE)RED(pixel);
  g = (BYTE)GREEN(pixel);
  b = (BYTE)BLUE(pixel);
  image->data[by][(x-1)*RGBSIZE]   = r;
  image->data[by][(x-1)*RGBSIZE+1] = g;
  image->data[by][(x-1)*RGBSIZE+2] = b;
}


/*-------------------------------------------------------------------*/
int GetPixel(IMAGE* image, ULONG x, ULONG y)
{
  int pixel;

  if (image->ModeType == BINMODE) pixel = GetBitPixel(image, x, y);
  else pixel = GetGrayPixel(image, x, y);
  return pixel;
}


void PutPixel(IMAGE* image, ULONG x, ULONG y, int pixel)
{
  if (image->ModeType == BINMODE) PutBitPixel(image, x, y, pixel);
  else PutGrayPixel(image, x, y, pixel);
}


/*-------------------------------------------------------------------*/
RGBPIXEL GetRGBPixelEx(IMAGE* image, ULONG x, ULONG y)
{
  RGBPIXEL  pixel;
  int       tmpInt;

  
  if(image->ModeType==BINMODE)
  {
    tmpInt = GetBitPixel(image, x, y);
    pixel = (tmpInt==BIN_BLACK) ? RGB_BLACK : RGB_WHITE;
  }
  else
    if(image->ModeType==GRAYMODE)
    {
      tmpInt = GetGrayPixel(image, x, y);
      pixel = RGB((BYTE)tmpInt, (BYTE)tmpInt, (BYTE)tmpInt);
    }
    else
      pixel = GetRGBPixel(image, x, y);

  return pixel;
}


void PutRGBPixelEx(IMAGE* image, ULONG x, ULONG y, RGBPIXEL pixel)
{
  int       tmpInt;

  if(image->ModeType==BINMODE)
  {
    tmpInt = (pixel==RGB_BLACK) ? BIN_BLACK : BIN_WHITE;
    PutBitPixel(image, x, y, tmpInt);
  }
  else
    if(image->ModeType==GRAYMODE)
    {
      tmpInt=RED(pixel);
      PutGrayPixel(image, x, y, tmpInt);
    }
    else
      PutRGBPixel(image, x, y, pixel);
}

/*===================================================================*/
/*=======================  General routines  ========================*/
/*===================================================================*/
YESNO SetImageSize(IMAGE* image, ULONG x, ULONG y)
{
  x=max(1,x); y=max(1,y);
  image->ImageSizeX = x;
  image->ImageSizeY = y;
  SetImageBounds(image, 1, 1, x, y);

  return (YES);
}


/*-------------------------------------------------------------------*/
YESNO SetDefaultColor(IMAGE* image, ULONG color)
{
  image->DefaultColor = color;
  return (YES);
}


/*-------------------------------------------------------------------*/
YESNO SetImageBounds(IMAGE* image, ULONG bx, ULONG by, ULONG sizex, ULONG sizey)
{
  ULONG ex,ey;

  if((sizex<1) || (sizey<1))
  {
    ErrorMessage(E_IMSIZE);
    return(NO);
  }
  ex = bx + sizex - 1;
  ey = by + sizey - 1;

  bx = max(bx,1);
  by = max(by,1);
  ex = min(ex,image->ImageSizeX);
  ey = min(ey,image->ImageSizeY);

  image->Bound.Left  = bx;
  image->Bound.Right = ex;
  image->Bound.Upper = by;
  image->Bound.Lower = ey;

  return (YES); /* values are rounded automaticly */
}


/*-------------------------------------------------------------------*/
void FlushImage(IMAGE* image)
{
  ULONG  n,rest,k,i;
  BYTE   *RData, *GData, *BData;


  if(image->FileMode == OUTPUT)
  {
    n=image->LastBufLine - image->FirstBufLine +1;
    rest=image->ImageSizeY - image->LastBufLine;

    WriteLines (image,n);

    if(rest>0)
    {
      EmptyLine(image,image->BufCount);
      switch(image->ModeType) 
      {
        case BINMODE:
          for(k=1;k<=rest;k++)
          {
            WritePbmBitLine(image->FilePointer, image->data[image->BufCount], image->ImageBytesX);
          }
          break;
        case GRAYMODE:  
          for(k=1;k<=rest;k++)
          {
            WritePgmLine(image->FilePointer, image->data[image->BufCount], image->ImageBytesX);
          }
          break;
        case RGBMODE: 
          RData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
          GData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
          BData = (BYTE*)allocate(sizeof(BYTE)*image->ImageSizeX);
          for(k=1;k<=rest;k++)
          {
            for(i=0;i<image->ImageSizeX;i++)
            {
              RData[i] = RED(image->data[image->BufCount][i*RGBSIZE]);
              GData[i] = GREEN(image->data[image->BufCount][i*RGBSIZE+1]);
              BData[i] = BLUE(image->data[image->BufCount][i*RGBSIZE+2]);
            }
            WritePpmLine(image->FilePointer, RData, GData, BData, image->ImageSizeX);
          }
          deallocate(RData);
          deallocate(GData);
          deallocate(BData);
          break;
      }
    }
    fflush (image->FilePointer);
  }
}


/*-------------------------------------------------------------------*/
YESNO ImageRewind(IMAGE* image)
{
  ULONG n;


  if(image->TargetType == DISK)
  {
    FlushImage(image);
    rewind(image->FilePointer);

    image->FileMode = INPUT;

    switch (image->ModeType) 
    {
      case BINMODE:
        ReadPbmHeader(image->FilePointer, &(image->ImageSizeX), &(image->ImageSizeY));
        break;
      case GRAYMODE:
        ReadPgmHeader(image->FilePointer, &(image->ImageSizeX), &(image->ImageSizeY), &(image->Max));
        break;
      case RGBMODE:
        ReadPpmHeader(image->FilePointer, &(image->ImageSizeX), &(image->ImageSizeY), &(image->Max));
        break;
     }
     image->ImageBytesX = BytesPerLine(image);
     image->Bound.Left  = 1;
     image->Bound.Right = image->ImageSizeX;
     image->Bound.Upper = 1;
     image->Bound.Lower = image->ImageSizeY;

     n=min(image->LinesInBuffer,image->ImageSizeY);

     image->FirstBufLine   = 1;
     image->LastBufLine    = n;
     image->BufCount       = 0;
     ReadLines(image, n);
     image->LinesReadTotal = n;
     image->FirstBufLine   = 1;
     image->LastBufLine    = n;
     image->BufCount       = 0;
     return (YES);
  }
  else 
    if(image->TargetType == MEMORY)
    {
       image->Bound.Left  = 1;
       image->Bound.Right = image->ImageSizeX;
       image->Bound.Upper = 1;
       image->Bound.Lower = image->ImageSizeY;
       image->BufCount       = 0;
       return(YES);
    }
    else return (NO);
}


/*===================================================================*/
/*======================= Image initialization ======================*/
/*===================================================================*/
YESNO ImageInit(char InputName[], IMAGE* image, FILEMODE mode, int AllowOverWrite, 
                BUFFERING Buffering, MODETYPE ModeType, TARGETTYPE TargetType, ULONG BufferSize)
{
  ULONG i,j,n;

  
  image->FilePointer = NULL;
  image->data        = NULL;
  image->ScrollSize  = 1;
  image->TargetType  = TargetType;
  image->ModeType    = ModeType;
  image->FileMode    = mode == INPUT ? INPUT : OUTPUT;
  image->Buffering   = Buffering;

  if(image->TargetType!=MEMORY)     // DISK or TEMPORARY access
  {
    if((image->FilePointer = FileOpen(InputName, mode, AllowOverWrite))==NULL)
    {
      return (NO);
    }

    if(image->FileMode == INPUT)
    {
      image->FileType = AnalyzeFileType(InputName);
      switch (image->FileType) 
      {
        case PBM:
          image->ModeType = BINMODE; 
          ReadPbmHeader(image->FilePointer, &(image->ImageSizeX), &(image->ImageSizeY));
          image->Max = 1;
          break;
        case PGM:
          image->ModeType = GRAYMODE; 
          ReadPgmHeader(image->FilePointer, &(image->ImageSizeX), &(image->ImageSizeY), &(image->Max));
          break;
        case PPM:
          image->ModeType = RGBMODE; 
          ReadPpmHeader(image->FilePointer, &(image->ImageSizeX), &(image->ImageSizeY), &(image->Max));
          break;
        default:  
          ErrorMessage("Image %s has not supported type or not exist",InputName);
          return(NO);
      }
    }
    else // OUTPUT
    {
      switch (image->ModeType) 
      {
        case BINMODE:
          image->FileType  = PBM; 
          image->Max       = 1;
          WritePbmHeader(image->FilePointer, image->ImageSizeX, image->ImageSizeY);
          break;
        case GRAYMODE:
          image->FileType  = PGM; 
          image->Max       = 255;
          WritePgmHeader(image->FilePointer, image->ImageSizeX, image->ImageSizeY, image->Max);
          break;
        case RGBMODE:
          image->FileType  = PPM;
          image->Max       = 255;
          WritePpmHeader(image->FilePointer, image->ImageSizeX, image->ImageSizeY, image->Max);
          break;
        default:
          return (NO);
      }
    }
  }
  else //MEMORY
  {
    image->Buffering   = WHOLEIMAGE;
    BufferSize = 0;
    switch (image->ModeType) 
    {
      case BINMODE:
        image->FileType  = PBM; 
        image->Max       = 1;
        break;
      case GRAYMODE:
        image->FileType  = PGM; 
        image->Max       = 255;
        break;
      case RGBMODE:
        image->FileType  = PPM;
        image->Max       = 255;
        break;
      default:
        return (NO);
    }
  }

  switch (image->ModeType) 
  {
    case BINMODE:
      image->DefaultColor = BIN_WHITE;
      break;
    case GRAYMODE:
      image->DefaultColor = 255;
      break;
    case RGBMODE:
      image->DefaultColor = RGB_WHITE;
      break;
  }

  if(BufferSize == 0) BufferSize = LINESAUTOMAX;

  switch(image->Buffering)
  {
    case WHOLEIMAGE:    
      image->LinesInBuffer  = image->ImageSizeY;
      break;
    case BUFACCESS:
      image->LinesInBuffer = min (image->ImageSizeY, BufferSize);
      break;
    default: /* case AUTOMATIC: */
      image->LinesInBuffer = min (image->ImageSizeY, BufferSize);
     break;
  }

  image->ImageBytesX = BytesPerLine(image);
  image->Bound.Left  = 1;
  image->Bound.Right = image->ImageSizeX;
  image->Bound.Upper = 1;
  image->Bound.Lower = image->ImageSizeY;


  /* --- allocate memory --- */
  if(image->LinesInBuffer >= LINESMAX)
  {
    ErrorMessage(E_LARGEBUFFER);
    fclose(image->FilePointer);
    return(NO);
  }

  if((image->data = (BYTE**)allocate((image->LinesInBuffer+1)*sizeof(BYTE*))) == NULL)
  {   
    ErrorMessage(E_NOMEMORY);
    fclose(image->FilePointer);
    return(NO);
  }
  for(i=0; i<image->LinesInBuffer; i++) image->data[i]=NULL;

  for(i=0; i<image->LinesInBuffer; i++)
  {
    image->data[i] = (BYTE*)allocate(image->ImageBytesX);
    if(image->data[i] == NULL)
    {
      if(i>=LINESMIN && image->Buffering==AUTOMATIC && image->FileMode==INPUT)
      {
        image->LinesInBuffer = i;
        image->Buffering=BUFACCESS; /* switch to buffering mode */
        break;
      }
      else
      {
        // deallocation
        for (j=0; j<i; j++) 
        {
          deallocate(image->data[j]);
          image->data[j]=NULL;
        }
        deallocate(image->data);
        image->data=NULL;
        fclose(image->FilePointer);
        return(NO);
      }
    }
  }

  // After all allocation succeeded, we can specify real target type for normal deinitialization
  /* for AUTOMATIC mode: if initialization done -> switch to BUFACCESS */
  if(image->Buffering == AUTOMATIC) image->Buffering = BUFACCESS;
  n=min(image->LinesInBuffer,image->ImageSizeY);

  image->LinesReadTotal = 0;
  image->FirstBufLine   = 1;
  image->LastBufLine    = n;
  image->BufCount       = 0;

  switch(image->FileMode)
  {
    case INPUT: /* --- fill buffer (for INPUT) --- */
      ReadLines(image, n);
      image->LinesReadTotal = n;
      image->FirstBufLine   = 1;
      image->LastBufLine    = n;
      image->BufCount       = 0;
      break;

    default: /* case OUTPUT: */
      for (i=0;i<n;i++)
      {
        EmptyLine(image, image->BufCount);
        image->BufCount++;
      }
      image->FirstBufLine   = 1;
      image->LastBufLine    = n;
      image->BufCount       = 0;
      image->LinesReadTotal = 0;
      break;
   }

  return (YES);
}



/*------------------------  Free Image  -------------------------*/
YESNO ImageDone(IMAGE* image)
{
  ULONG   y;

  if(image==NULL) return (YES);

  switch (image->TargetType)
  {
    case DISK:      
      FlushImage(image);
      if(image->FilePointer != NULL) fclose(image->FilePointer);
      break;
    case MEMORY:
      break;
    case TEMPORARY:      
      if(image->FilePointer != NULL) fclose(image->FilePointer);
      break;
    default:
      break;
  }

  /* --- deallocate memory --- */
  if(image->data != NULL)
  {
    for(y=0; y<image->LinesInBuffer; y++)
    {
      if(image->data[y] != NULL) deallocate(image->data[y]);
    }
    deallocate(image->data);
  }

  return(YES);
}


/*------------------------ Export Image -------------------------*/
YESNO ExportImage(IMAGE* image, BYTE* target, ULONG Color)
{
  ULONG     x,y,i;
  BYTE      pixel,r,g,b;


  if(image->Buffering != WHOLEIMAGE) return NO;

  if(Color!=0)
  {
    r = (BYTE)Color&0xFF;  g = (BYTE)(Color>>8)&0xFF;  b = (BYTE)(Color>>16)&0xFF;
  }
  else
  {
    r = 1;  g = 1;  b = 1;
  }

  SetImageBounds(image, 1, 1, image->ImageSizeX, image->ImageSizeY);
  i=0;
  for(y=1; y<=image->ImageSizeY; y++)
  {
    for(x=1; x<=image->ImageSizeX; x++)
    {
      switch (image->ModeType)
      {
        case BINMODE:
          pixel = GetBitPixel(image, x, y);
          if(pixel!=0)
          {
            target[i*3]   = r;
            target[i*3+1] = g;
            target[i*3+2] = b;
          }
          break;
        case GRAYMODE:
          pixel = GetGrayPixel(image, x, y);
          target[i*3]   = pixel;
          target[i*3+1] = pixel;
          target[i*3+2] = pixel;
          break;
        case RGBMODE:
          memcpy(&(target[i*3]), &(image->data[y][(x-1)*RGBSIZE]),3);
          break;
      }
      i++;
    }
  }
  return (YES);
}


/*------------------------ Export Image -------------------------*/
YESNO ExportImageEx(IMAGE* image, BYTE* target, ULONG Color, ULONG BackColor)
{
  ULONG x,y,i;
  BYTE pixel,r,g,b,br,bg,bb;


  if(image->Buffering != WHOLEIMAGE) return NO;

  if(Color!=0)
  {
    r = (BYTE)Color&0xFF;  g = (BYTE)(Color>>8)&0xFF;  b = (BYTE)(Color>>16)&0xFF;
  }
  else
  {
    r = 1;  g = 1;  b = 1;
  }

  if (BackColor!=0)
  {
    br = (BYTE)BackColor&0xFF;  bg = (BYTE)(BackColor>>8)&0xFF;  bb = (BYTE)(BackColor>>16)&0xFF;
  }
  else
  {
    br = 255;  bg = 255;  bb = 255;
  }

  SetImageBounds(image, 1, 1, image->ImageSizeX, image->ImageSizeY);
  i=0;
  for(y=1; y<=image->ImageSizeY; y++)
  {
    for(x=1; x<=image->ImageSizeX; x++)
    {
      switch (image->ModeType)
      {
        case BINMODE:
          pixel = GetBitPixel(image, x, y);
          if(pixel!=0)
          {
            target[i*3]   = r;
            target[i*3+1] = g;
            target[i*3+2] = b;
          }
          else
          {
            target[i*3]   = br;
            target[i*3+1] = bg;
            target[i*3+2] = bb;
          }
          break;
        case GRAYMODE:
          pixel = GetGrayPixel(image, x, y);
          target[i*3]   = pixel;
          target[i*3+1] = pixel;
          target[i*3+2] = pixel;
          break;
        case RGBMODE:
          memcpy(&(target[i*3]), &(image->data[y][(x-1)*RGBSIZE]),3);
          break;
      }
      i++;
    }
  }
  return (YES);
}


YESNO ExportImageBin(IMAGE* image, BYTE* target, int plane)
{
  ULONG     x,y,i;
  BYTE      pixel;


  if(image->Buffering != WHOLEIMAGE) return NO;

  SetImageBounds(image, 1, 1, image->ImageSizeX, image->ImageSizeY);
  i=0;
  for(y=1; y<=image->ImageSizeY; y++)
  {
    for(x=1; x<=image->ImageSizeX; x++)
    {
      if (image->ModeType==BINMODE)
      {
        pixel = GetBitPixel(image, x, y);
        target[i] = target[i] | (pixel<<plane);
        target[i+1] = target[i];
        target[i+2] = target[i];
      }
      i+=3;
    }
  }
  return (YES);
}


/*------------------------ Dump Image -------------------------*/
YESNO DumpImage(IMAGE* image, char *Name)
{
  IMAGE     Dump;
  ULONG     x,y;
  RGBPIXEL  pixel;


  SetImageSize(&Dump, image->ImageSizeX, image->ImageSizeY);
  ImageInit(Name, &Dump, OUTPUT, YES, AUTOMATIC, image->ModeType, DISK, 10);

  ImageRewind(image);
  for(y=1; y<=image->ImageSizeY; y++)
  {
    for(x=1; x<=image->ImageSizeX; x++)
    {
      pixel=GetRGBPixelEx(image, x, y);
      PutRGBPixelEx(&Dump, x, y, pixel);
    }
  }
  ImageDone(&Dump);
  return (YES);
}

/*------------------------ Copy Image -------------------------*/
YESNO CopyImage(IMAGE* dest, IMAGE* src, char* Name, TARGETTYPE TargetType)
{
  ULONG     x,y;
  RGBPIXEL  pixel;


  SetImageSize(dest, src->ImageSizeX, src->ImageSizeY);
  ImageInit(Name, dest, src->FileMode, YES, src->Buffering, src->ModeType, TargetType, 10);

  ImageRewind(src);
  for(y=1; y<=src->ImageSizeY; y++)
  {
    for(x=1; x<=src->ImageSizeX; x++)
    {
      pixel=GetRGBPixelEx(src, x, y);
      PutRGBPixelEx(dest, x, y, pixel);
    }
  }
  return (YES);
}


