/*-------------------------------------------------------------------*/
/* Mapseg.c          Pavel Kopylov                                   */
/*                                                                   */
/* - Color segmentation + Bit level segmentation                     */
/* - Image decomposition                                             */
/* - Backgroung analyzing                                            */
/* - Color map saving                                                */
/*-------------------------------------------------------------------*/

#define ProgName       "MAPSEG"
#define VersionNumber  "0.15"
#define LastUpdated    "22.03.2004"

#include <math.h>
#include <assert.h>
#include "mapseg.h"

#include "file.h"
#include "image.h"
#include "pgm.h"
#include "memctrl.h"
#include "interfc.h"
#include "project.h"
#include "colorfilt.h"

// RLS from Ismo
#include "rls.h"
#include "distance.h"
#include "criteria.h"
#include "iter.h"
#include "cb_util.h"

// Colors
#include "colordis.h"
#include "colorfilt.h"

//Shape
#include "shape.h"


ULONG     NodesAdded, Colors;
#define DELTA 0.45


void     InsertNodeToTrainingSet(TRAININGSET* TS, BOOKNODE* node, int Count, int tweak);
void     QuantizeImage(IMAGE* InputImage, IMAGE* QuantizedImage, TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA, MULTITREEm* MT);
double   GenerateCodebook(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA, SEGMPARAMS *Params, MULTITREEm *MT);
ULONG    CalculateLayer(char *LayerName, int ContextSize);
int      BinToGrayCode(int value);
int      GrayToBinCode(int value);
YESNO    CheckExt(char* FileName, char* Extention);
MULTITREEm *NewMTreeNode();
ULONG ReadRAWToColorTree(MULTITREEm **MT, BYTE *Data, ULONG XSize, ULONG YSize);
int GetExtraProperty(IMAGE* image, ULONG x, ULONG y);

/*-------------------------------------------------------------------*/
YESNO SaveColorMap(PROJECTDATA *Project)
{
  TRAININGSET     TS;
  IMAGE           InputImage;
  FILE            *f;
  ULONG           NColors, i;
  MULTITREEm      *MT;

  
  ImageInit(Project->InputName, &InputImage, INPUT, NO, AUTOMATIC, BINMODE, DISK, 10);
  NColors = ConstructColorTree(&MT, &InputImage, NULL, 2, 0, 5);

  // Reserve the memory for the Training Set
  InitializeTrainingSet(&TS, NColors, InputImage.Max, 3);

  // Proceed the tree & put everything to training set
  ConvertTreeToTrainingSet(&TS, MT, NColors, 0);

  f = fopen(Project->OutputName,"wt");
  if (f==NULL)
  {
    DeallocateMTree(MT);
    FreeCodebook(&TS);
    ImageDone(&InputImage);
    return NO;
  }

  fprintf(f,"%d\n",NColors);
  for(i=0;i<NColors;i++)
  {
    fprintf(f,"%d %d %d\n",VectorScalar(&TS,i,0), VectorScalar(&TS,i,1), VectorScalar(&TS,i,2));
  }
  fclose(f);

  DeallocateMTree(MT);
  FreeCodebook(&TS);
  ImageDone(&InputImage);
  return YES;
}


/*-------------------------------------------------------------------*/
YESNO mapseg(PROJECTDATA *Project, SEGMPARAMS *Params)
{
  TRAININGSET     TS;
  CODEBOOK        CB;
  PARTITIONING    PA;
  IMAGE           InputImage, LayerImage, QuantizedImage;
  int             ret=0, Layers;
  char            LayerName[127] = {'\0'}, tmpStr[127] = {'\0'};
  ULONG           NColors;
  MULTITREEm       *MT;


  // Read an image into Binary tree
  // We are don't care about image type now, because it will be set inside
  CheckFileName(Project->OutputName, FormatNamePPM);

  Layers = Project->Layers;
  if(Layers==1)
  {
    strcpy(LayerName, Project->ProjectParams[0].LayerFileName);
    Project->Layers = 0;
    deallocate(Project->ProjectParams);
    Project->ProjectParams=NULL;

    ImageInit(Project->InputName, &InputImage, INPUT, NO, AUTOMATIC, BINMODE, DISK, 10);
    ImageInit(LayerName, &LayerImage, INPUT, NO, AUTOMATIC, BINMODE, DISK, 10);
    NColors = ConstructColorTree(&MT, &InputImage, &LayerImage, Params->QuietLevel, Params->tweak, Params->Threshold);
    ImageDone(&LayerImage);
    ImageDone(&InputImage);
  }
  else
  {
    ImageInit(Project->InputName, &InputImage, INPUT, NO, AUTOMATIC, BINMODE, DISK, 10);
    NColors = ConstructColorTree(&MT, &InputImage, NULL, Params->QuietLevel, Params->tweak, Params->Threshold);
    ImageDone(&InputImage);
  }

  if(Params->QuietLevel >= 2) PrintMessage("Training set contains now %u cells.\n", NColors);

  // Reserve the memory for the Training Set
  InitializeTrainingSet(&TS, NColors, InputImage.Max, Params->tweak);

  // Proceed the tree & put everything to training set
  ConvertTreeToTrainingSet(&TS, MT, NColors, Params->tweak);

  if (Params->QuietLevel>3) PrintCodeBook(&TS, 2);

  /*ReadTrainingSet("ts.aaa", &TS);*/

  /////////////////////////////////////////////////////////////
  // Processing part
  /////////////////////////////////////////////////////////////
  ImageInit(Project->InputName, &InputImage, INPUT, NO, AUTOMATIC, BINMODE, DISK, 10);


  if (Params->Separation || BookSize(&TS)<=Params->Colors) 
  {
    CreateNewCodebook(&CB, BookSize(&TS), &TS);
    CreateNewPartitioning(&PA, &TS, BookSize(&TS));
    CopyCodebook(&TS, &CB);
    ColorSeparation(&InputImage, Project, Params, &TS, &CB, &PA, MT);
    FreePartitioning(&PA);
    FreeCodebook(&CB);
    ret=2;
  }
  else
  { 
    // Reducing the amount of colors
    GenerateCodebook(&TS, &CB, &PA, Params, MT);                        
    
    //WriteCodebook("cb.aaa", &CB, 1);
    //ReadCodebook("cb.aaa", &CB);
    if (Params->QuietLevel>3) PrintCodeBook(&CB, Params->tweak);
    
    if (Params->CI==CR_FTest)
    {
      NColors = ReadVectors(&CB, Params->tweak);
      if(Params->QuietLevel>1) PrintMessage("\tReal colors:\t%u\n", NColors);
    }
    else // Quantize Image
    {
      SetImageSize(&QuantizedImage, InputImage.ImageSizeX, InputImage.ImageSizeY);
      ImageInit(Project->OutputName, &QuantizedImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);
      QuantizeImage(&InputImage, &QuantizedImage, &TS, &CB, &PA, MT);
      ImageDone(&QuantizedImage);
    } 
    FreePartitioning(&PA);
    FreeCodebook(&CB);
    ret=1;
  }

  ImageDone(&InputImage);
  FreeCodebook(&TS);
  DeallocateMTree(MT);

  return ret;
}

 
/*-------------------------------------------------------------------*/
YESNO mapreconst(PROJECTDATA *Project)
{
  CODEBOOK        CB;
  IMAGE           ReconstructedImage;
  IMAGE           *InputImage;
  UINT32          i, k;
  int             index;
  RGBPIXEL        color, background;
  BYTE            bw;
  YESNO           found, BKSet;
  ULONG           x, y;


  if (Project->Colors==0) return NO;

  if (Project->BackGround==0) BKSet=NO;
  else
  {
    BKSet=YES;
    background = Project->BackGround;
  }

  CreateNewTrainingSet(&CB, Project->Layers, 3, 1, 1, 0, 255, 0, "MAPReconst");
  BookSize(&CB) = Project->Colors;

  // Prepare input and output images
  InputImage = (IMAGE*)allocate(Project->Layers*sizeof(IMAGE));

  for(i=0; i<Project->Layers; i++)
  {
    CheckFileName(Project->ProjectParams[i].LayerFileName, FormatNamePBM);
    ImageInit(Project->ProjectParams[i].LayerFileName, &(InputImage[i]), INPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);

    for(k=0; k<Project->Colors; k++)
    {
      if(Project->ProjectParams[i].LayerIndex == Project->Color[k].LayerIndex)
      {
        VectorScalar(&CB,i,0) = RED(Project->Color[k].Color);
        VectorScalar(&CB,i,1) = GREEN(Project->Color[k].Color);
        VectorScalar(&CB,i,2) = BLUE(Project->Color[k].Color);
      }
    }
  }

  SetImageSize(&ReconstructedImage, InputImage[0].ImageSizeX, InputImage[0].ImageSizeY);
  ImageInit(Project->OutputName, &ReconstructedImage, OUTPUT, YES, AUTOMATIC, RGBMODE, DISK, 10);

  SetProgressVisibility(1);
  OpenProgressWindow();

  // Write the output image line by line
  for(y=1; y<=ReconstructedImage.ImageSizeY; y++ )
  {
    // Perform color separation for the current line of input image
    for( x=1; x<=ReconstructedImage.ImageSizeX; x++ )
    {
      index=0;
      found=NO;
      //for(i=0; i<Project->Layers && !index; i++)
      for(i=0; i<Project->Layers; i++)
      {
        bw = GetBitPixel(&(InputImage[i]), x, y);
        if(bw != 0)
        { 
          found=YES;
          index=i;
        }
      }
      if (found)
      {
        color = GetColorByIndex(&CB, index);
        PutRGBPixel(&ReconstructedImage, x, y, color);
      }
      else PutRGBPixel(&ReconstructedImage, x, y, background);
    }
    UpdateProgressBar(y, ReconstructedImage.ImageSizeY);
  }
  CloseProgressWindow();


  for(i=0; i<Project->Layers; i++)
  {
    ImageDone(&(InputImage[i]));
  }

  ImageDone(&ReconstructedImage);

  deallocate(InputImage);

  FreeCodebook(&CB);

  return YES;
}


/*-------------------------------------------------------------------*/
YESNO AnalyzeBackground(PROJECTDATA *Project)
{
  ULONG           *Array, i,max,maxpos;


  Array=(ULONG*)allocate(sizeof(ULONG)*Project->Layers);

  for(i=0;i<Project->Layers;i++)
  {
    Array[i] = CalculateLayer(Project->ProjectParams[i].LayerFileName, 10);
  }

  max=0; maxpos=0;
  for(i=0;i<Project->Layers;i++)
  {
    if(Array[i]>max)
    {
      max = Array[i];
      maxpos = i;
    }
  }

  Project->BackGround = Project->Color[maxpos].Color;
  remove(Project->ProjectParams[maxpos].LayerFileName);
  if(max!=0) DeleteItem(Project, maxpos);

  deallocate(Array);
  return YES;
}


/*-------------------------------------------------------------------*/
YESNO bitseg(PROJECTDATA *Project, SEGMPARAMS *Params)
{
  YESNO           ret;
  IMAGE           InputImage;
  ULONG           i;
  char            tmpStr[127];

  // We can only work with PGN's - GrayScales only
  ret = CheckExt(Project->InputName, FormatNamePGM);
  RETNO(ret);

  ImageInit(Project->InputName, &InputImage, INPUT, NO, AUTOMATIC, GRAYMODE, DISK, 10);
  BitSeparation(&InputImage, Project->OutputName, Params->GrayCode);
  
  Project->Layers = 8;
  Project->XSize = InputImage.ImageSizeX;
  Project->YSize = InputImage.ImageSizeY;
  
  Project->ProjectParams = (PROJECTPARAMETERS *)allocate(sizeof(PROJECTPARAMETERS)*Project->Layers);
  if (Project->ProjectParams == NULL) return NO;
  if (Params->Trees)
  {
    Project->Trees = Project->Layers;
    Project->Tree = (TREEPARAMETERS *)allocate(sizeof(TREEPARAMETERS)*Project->Trees);
    if (Project->Tree == NULL) return NO;
  }

  for(i=0; i<Project->Layers; i++)
  {
    //Generate Layer names
    PickFileName(Project->OutputName, tmpStr);
    sprintf(tmpStr, "%s-%i", tmpStr, i+1);
    CheckFileName(tmpStr, FormatNamePBM);
    Project->ProjectParams[i].LayerIndex = i;
    strcpy(Project->ProjectParams[i].LayerFileName, tmpStr);
    Project->ProjectParams[i].LayerType = 0;
    Project->ProjectParams[i].LayerDependency = i;

    if (Params->Trees)
    {
      PickFileName(Project->OutputName, tmpStr);
      sprintf(tmpStr, "%s-%i", tmpStr, i+1);
      CheckFileName(tmpStr, "ct");
      Project->Tree[i].LayerIndex = i;
      strcpy(Project->Tree[i].LayerFileName, tmpStr);
    }
  }
  ImageDone(&InputImage);

  return YES;
}


/*-------------------------------------------------------------------*/
YESNO bitreconst(PROJECTDATA *Project, YESNO GrayCode)
{
  IMAGE           ReconstructedImage;
  IMAGE           *InputImage;
  UINT32          i;
  int             pixel, plane, impixel;
  ULONG           x, y;


  if (Project->Colors>0) return NO;

  // Prepare input and output images
  InputImage = (IMAGE*)allocate(Project->Layers*sizeof(IMAGE));

  for(i=0; i<Project->Layers; i++)
  {
    CheckFileName(Project->ProjectParams[i].LayerFileName, FormatNamePBM);
    ImageInit(Project->ProjectParams[i].LayerFileName, &(InputImage[i]), INPUT, YES, AUTOMATIC, BINMODE, DISK, 10);
  }

  SetImageSize(&ReconstructedImage, InputImage[0].ImageSizeX, InputImage[0].ImageSizeY);
  ImageInit(Project->OutputName, &ReconstructedImage, OUTPUT, YES, AUTOMATIC, GRAYMODE, DISK, 10);

  SetProgressVisibility(1);
  OpenProgressWindow();

  for(y=1; y<=ReconstructedImage.ImageSizeY; y++ )
  {
    for( x=1; x<=ReconstructedImage.ImageSizeX; x++ )
    {
      pixel = 0;
      for (plane=Project->Layers-1; plane>=0; plane--)
      {
        impixel = GetBitPixel(&InputImage[plane], x, y);
        pixel = (pixel<<1) + impixel;
      }
      if (GrayCode) pixel = GrayToBinCode(pixel);
      PutGrayPixel(&ReconstructedImage, x, y, pixel);
    }
    UpdateProgressBar(y, ReconstructedImage.ImageSizeY);
  }
  CloseProgressWindow();

  for(i=0; i<Project->Layers; i++)
  {
    ImageDone(&(InputImage[i]));
  }
  ImageDone(&ReconstructedImage);
  deallocate(InputImage);

  return YES;
}


/*-------------------------------------------------------------------*/
/*-- Separation -----------------------------------------------------*/
/*-------------------------------------------------------------------*/
void ColorSeparation(IMAGE* QuantizedImage, PROJECTDATA *Project, SEGMPARAMS *Params, TRAININGSET* TS, 
                     CODEBOOK* CB, PARTITIONING* PA, MULTITREEm* MT)
{
  IMAGE     *OutputImage;
  char      tmpStr[127];
  RGBPIXEL  color;
  int       index, r,g,b;
  ULONG     x,y,i;
 

  Project->Layers = BookSize(CB);
  Project->Colors = Project->Layers;
  Project->XSize  = QuantizedImage->ImageSizeX;
  Project->YSize  = QuantizedImage->ImageSizeY;

  Project->ProjectParams = (PROJECTPARAMETERS *)allocate(sizeof(PROJECTPARAMETERS)*Project->Layers);
  assert(Project->ProjectParams != NULL);

  if (Params->Trees)
  {
    Project->Trees = Project->Layers;
    Project->Tree = (TREEPARAMETERS *)allocate(sizeof(TREEPARAMETERS)*Project->Trees);
    assert(Project->Tree != NULL);
  }
  Project->Color = (COLORPARAMETERS *)allocate(sizeof(COLORPARAMETERS)*Project->Colors);
  assert(Project->Color != NULL);

  // Prepare input and output images
  OutputImage = (IMAGE*)allocate(Project->Layers*sizeof(IMAGE));
  assert(OutputImage != NULL);

  for(i=0; i<Project->Layers; i++)
  {
    //Generate Layer names
    PickFileName(Project->OutputName, tmpStr);
    sprintf(tmpStr, "%s-%i", tmpStr, i+1);
    CheckFileName(tmpStr, FormatNamePBM);
    Project->ProjectParams[i].LayerIndex = i;
    strcpy(Project->ProjectParams[i].LayerFileName, tmpStr);
    Project->ProjectParams[i].LayerType = 0;
    Project->ProjectParams[i].LayerDependency = i;

    /* Initialize output image files and data strucutres */
    SetImageSize(&(OutputImage[i]), QuantizedImage->ImageSizeX, QuantizedImage->ImageSizeY);
    ImageInit(Project->ProjectParams[i].LayerFileName, &(OutputImage[i]), OUTPUT, YES, WHOLEIMAGE, BINMODE, DISK, 10);

    if (Params->Trees)
    {
      PickFileName(Project->OutputName, tmpStr);
      sprintf(tmpStr, "%s-%i", tmpStr, i+1);
      CheckFileName(tmpStr, "ct");
      Project->Tree[i].LayerIndex = i;
      strcpy(Project->Tree[i].LayerFileName, tmpStr);
    }
    Project->Color[i].LayerIndex = i;
    Project->Color[i].Color = RGB(VectorScalar(CB, i, 0), VectorScalar(CB, i, 1), VectorScalar(CB, i, 2));
  }

  //--- Preparations ok

  ImageRewind(QuantizedImage);
  SetProgressVisibility(1);
  OpenProgressWindow();

  // Analyze the input image line by line
  for( y=1; y<=QuantizedImage->ImageSizeY; y++ )
  {
    // Perform color separation for the current line of input image
    for( x=1; x<=QuantizedImage->ImageSizeX; x++ )
    {
      color = GetRGBPixelEx(QuantizedImage, x, y);
      r = (int)RED(color);
      g = (int)GREEN(color);
      b = (int)BLUE(color);

      if (MT->Children[r]!=NULL && MT->Children[r]->Children[g]!=NULL && MT->Children[r]->Children[g]->Children[b]!=NULL)
      {
        index = MT->Children[r]->Children[g]->Children[b]->index;
      }
      PutBitPixel(&(OutputImage[index]), x, y, 1);
    }
    UpdateProgressBar(y, QuantizedImage->ImageSizeY);
  }
  CloseProgressWindow();
  
  for(i=0; i<Project->Layers; i++)
  {
    ImageDone(&(OutputImage[i]));
  }
  deallocate(OutputImage);
}


/*-------------------------------------------------------------------*/
void ColorSeparationEx(IMAGE* QuantizedImage, IMAGE **ImageArray, TRAININGSET* TS, CODEBOOK* CB, 
                       PARTITIONING* PA, MULTITREEm* MT)
{
  IMAGE     *OutputImage;
  RGBPIXEL  color;
  int       index,i,r,g,b;
  ULONG     x,y;

 
  // Prepare input and output images
  OutputImage = (IMAGE*)allocate(BookSize(CB)*sizeof(IMAGE));

  for(i=0; i<BookSize(CB); i++)
  {
    /* Initialize output image files and data strucutres */
    SetImageSize(&(OutputImage[i]), QuantizedImage->ImageSizeX, QuantizedImage->ImageSizeY);
    ImageInit("", &(OutputImage[i]), OUTPUT, YES, WHOLEIMAGE, BINMODE, MEMORY, 10);
  }
  //--- Preparations ok

  ImageRewind(QuantizedImage);
  SetProgressVisibility(1);
  OpenProgressWindow();

  // Analyze the input image line by line
  for( y=1; y<=QuantizedImage->ImageSizeY; y++ )
  {
    // Perform color separation for the current line of input image
    for( x=1; x<=QuantizedImage->ImageSizeX; x++ )
    {
      color = GetRGBPixelEx(QuantizedImage, x, y);
      r = (int)RED(color);
      g = (int)GREEN(color);
      b = (int)BLUE(color);

      if (MT->Children[r]!=NULL && MT->Children[r]->Children[g]!=NULL && MT->Children[r]->Children[g]->Children[b]!=NULL)
      {
        index = MT->Children[r]->Children[g]->Children[b]->index;
      }
      PutBitPixel(&(OutputImage[index]), x, y, 1);
    }
    UpdateProgressBar(y, QuantizedImage->ImageSizeY);
  }
  CloseProgressWindow();
  
  (*ImageArray)=OutputImage;
}


/*-------------------------------------------------------------------*/
void BitSeparation(IMAGE* QuantizedImage, char *OutputName, YESNO GrayCode)
{
  IMAGE     *OutputImage;
  char      OutputFN[127], tmpStr[127];
  int       color, pmask, plane, pel, bit;
  ULONG     x,y;

 
  // Prepare input and output images
  OutputImage = (IMAGE*)allocate(8*sizeof(IMAGE));

  for(plane=0; plane<8; plane++)
  {
    // Generate name for output files
    PickFileName(OutputName, tmpStr);
    sprintf(OutputFN, "%s-%i", tmpStr, plane+1);
    CheckFileName(OutputFN, FormatNamePBM);

    /* Initialize output image files and data strucutres */
    SetImageSize(&(OutputImage[plane]), QuantizedImage->ImageSizeX, QuantizedImage->ImageSizeY);
    ImageInit(OutputFN, &(OutputImage[plane]), OUTPUT, YES, WHOLEIMAGE, BINMODE, DISK, 10);
  }
  //--- Preparations ok

  ImageRewind(QuantizedImage);
  SetProgressVisibility(1);
  OpenProgressWindow();

  // Analyze the input image line by line
  for( y=1; y<=QuantizedImage->ImageSizeY; y++ )
  {
    // Perform color separation for the current line of input image
    for( x=1; x<=QuantizedImage->ImageSizeX; x++ )
    {
      color = GetGrayPixel(QuantizedImage, x, y);
      if (GrayCode) color = BinToGrayCode(color);
      for (plane=7; plane>=0; plane--)
      {
        pmask=1<<(plane);
        pel = (color & pmask) ? 1 : 0;
        bit = (pel != 0) ? 1 : 0;
        PutBitPixel(&(OutputImage[plane]), x, y, bit);
      }
    }
    UpdateProgressBar(y, QuantizedImage->ImageSizeY);
  }
  CloseProgressWindow();
  
  for(plane=0; plane<8; plane++)
  {
    ImageDone(&(OutputImage[plane]));
  }
  deallocate(OutputImage);
}


/*-------------------------------------------------------------------*/
void BitSeparationEx(IMAGE* QuantizedImage, IMAGE **ImageArray, YESNO GrayCode)
{
  IMAGE     *OutputImage;
  int       color, pmask, plane, pel, bit;
  ULONG     x,y;

 
  // Prepare input and output images
  OutputImage = (IMAGE*)allocate(8*sizeof(IMAGE));

  for(plane=0; plane<8; plane++)
  {
    /* Initialize output image files and data strucutres */
    SetImageSize(&(OutputImage[plane]), QuantizedImage->ImageSizeX, QuantizedImage->ImageSizeY);
    ImageInit("",&(OutputImage[plane]), OUTPUT, YES, WHOLEIMAGE, BINMODE, MEMORY, 10);
  }
  //--- Preparations ok

  ImageRewind(QuantizedImage);
  SetProgressVisibility(1);
  OpenProgressWindow();

  // Analyze the input image line by line
  for( y=1; y<=QuantizedImage->ImageSizeY; y++ )
  {
    // Perform color separation for the current line of input image
    for( x=1; x<=QuantizedImage->ImageSizeX; x++ )
    {
      color = GetGrayPixel(QuantizedImage, x, y);
      if (GrayCode) color = BinToGrayCode(color);
      for (plane=7; plane>=0; plane--)
      {
        pmask=1<<(plane);
        pel = (color & pmask) ? 1 : 0;
        bit = (pel != 0) ? 1 : 0;
        PutBitPixel(&(OutputImage[plane]), x, y, bit);
      }
    }
    UpdateProgressBar(y, QuantizedImage->ImageSizeY);
  }
  CloseProgressWindow();
  
  (*ImageArray)=OutputImage;
}


/*-------------------------------------------------------------------*/
/*-- Service functions ----------------------------------------------*/
/*-------------------------------------------------------------------*/
void QuantizeImage(IMAGE* InputImage, IMAGE* QuantizedImage, TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA, MULTITREEm* MT)
{
  RGBPIXEL  color, color2, color1;
  int       index;
  ULONG     x, y;
  BYTE      r, g, b;


  ImageRewind(InputImage);
  SetProgressVisibility(1);
  OpenProgressWindow();

  for(y=1; y<=InputImage->ImageSizeY; y++ )
  {
    // Perform color separation for the current line of input image
    for( x=1; x<=InputImage->ImageSizeX; x++ )
    {
      color = GetRGBPixelEx(InputImage, x, y);
//      color1 = LocalMedian(InputImage, x, y, 0, 8);
      index = GetColorIndex(TS, CB, PA, MT, color, color1);
      r = VectorScalar(CB, index, 0);
      g = VectorScalar(CB, index, 1);
      b = VectorScalar(CB, index, 2);
      color2 = RGB(r, g, b);
      PutRGBPixel(QuantizedImage, x, y, color2);
    }
    UpdateProgressBar(y, InputImage->ImageSizeY);
  }
  CloseProgressWindow();
}


/*-------------------------------------------------------------------*/
void PrintCodeBook(CODEBOOK* CB, int mode)
{
  int        i;
  VECTORTYPE v;
  double     L,A,B,X,Y,Z;

  
  PrintMessage("CodeBook contains following colors:\n\n");
  switch (mode)
  {
  case 0:
    PrintMessage("Count\tFreq\t  R   G   B\n-----------------------------\n");
    for(i=0; i<BookSize(CB); i++)
    {
      PrintMessage("%d\t%d\t", i, Node(CB, i).freq); PrintVector(Vector(CB,i), 3, 1); PrintMessage("\n");
    }
    break;

  case 1:
    PrintMessage("Count\tFreq\t  R   G   B   L\n---------------------------------\n");
    for(i=0; i<BookSize(CB); i++)
    {
      PrintMessage("%d\t%d\t", i, Node(CB, i).freq); PrintVector(Vector(CB,i), 4, 1); PrintMessage("\n");
    }
    break;

  case 2:
    PrintMessage("Count\tFreq\t  R   G   B     L*      a*       b*\n---------------------------------------------------\n");
    for(i=0; i<BookSize(CB); i++)
    {
      v = Vector(CB,i);

      RGB_to_XYZ((BYTE)v[0],(BYTE)v[1],(BYTE)v[2], &X, &Y, &Z);
      XYZ_to_LAB(D65, X, Y, Z, &L, &A, &B);

      //if(v[2]<5)
      {
        PrintMessage("%d\t%d\t", i, Node(CB, i).freq); PrintVector(v, 3, 1);
        PrintMessage("\t%3.4f\t%3.4f\t%3.4f\n", L, A, B);
      }
    }
    break;
  }


  PrintMessage("\n");
}

/* =========================  P R O C E S S  ======================= */
void InitializeTrainingSet(TRAININGSET* TS, int size, int maxvalue, int tweak)
{
  char GenMethod[MaxGenMethodLength];
  
  strcpy(GenMethod, ProgName);
  strcat(GenMethod, " ");
  strcat(GenMethod, VersionNumber);
  if (tweak) CreateNewTrainingSet(TS, size, 4, 1, 1, 0, maxvalue, 0, GenMethod);
  else CreateNewTrainingSet(TS, size, 3, 1, 1, 0, maxvalue, 0, GenMethod);
  BookSize(TS) = 0;
}


/*-------------------------------------------------------------------*/
int CompareNodes(void* a, void* b, void* info, int tweak)
{
  int  ret;

  if(tweak)
    ret = CompareVectors( ((BOOKNODE*) a)->vector,((BOOKNODE*) b)->vector, 4);
  else
    ret = CompareVectors( ((BOOKNODE*) a)->vector,((BOOKNODE*) b)->vector, 3);

  return ret;
}    


ULONG ReadVectors(CODEBOOK* CB, int tweak)
{
  CODEBOOK    CB1;
  VECTORTYPE  v, *va;
  int         i,found;
  ULONG       ret, j;

  
  if(!tweak) ret=BookSize(CB);
  else 
  {
    CreateNewCodebook(&CB1, BookSize(CB), CB);
    CopyCodebook(CB, &CB1);
    va = (VECTORTYPE*)allocate(sizeof(VECTORTYPE)*BookSize(&CB1));
  
    ret=0;
    for(i=0; i<BookSize(&CB1); i++)
    {
      v = Vector(&CB1, i);
      found=0;
      for(j=0;j<ret;j++)
      {
        if (va[j][0]==v[0] && va[j][1]==v[1] && va[j][2]==v[2]) found=1;
      }
      if (!found)
      {
        va[ret] = CreateEmptyVector(3);
        va[ret][0]=v[0]; va[ret][1]=v[1]; va[ret][2]=v[2];
        ret++;
      }
    }
    for(j=0; j<ret; j++) FreeVector(va[j]);
    deallocate(va);
    FreeCodebook(&CB1);
  }

  return ret;
}


/*-------------------------------------------------------------------*/
void ConvertTreeToTrainingSet(TRAININGSET* TS, MULTITREEm* MT, int NColors, int tweak)
{
  BOOKNODE  node;
  int       Count, r, g, b;
  MULTITREEm  *T;


  IncreaseCodebookSize(TS, NColors);
  Count=0;
  if (tweak) node = CreateEmptyNode(4);
  else node = CreateEmptyNode(3);

  for (r=0;r<256;r++)
  {
    if (MT->Children[r] != NULL)
    for(g=0;g<256;g++)
    {
      if (MT->Children[r]->Children[g] != NULL)
      for(b=0;b<256;b++)
      {
        T=MT->Children[r]->Children[g]->Children[b];
        if (T!=NULL) 
        {
          node.freq = T->freq;
          node.vector[0]=r;
          node.vector[1]=g;
          node.vector[2]=b;

          T->index = Count;

          if (tweak) CopyNode(&node, &Node(TS, Count), 4);
          else CopyNode(&node, &Node(TS, Count), 3);
          TotalFreq(TS) += node.freq;
          Count++;
        }
      }
    }
  }
  FreeNode(node);
}


/*-------------------------------------------------------------------*/
double GenerateCodebook(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA, SEGMPARAMS *Params, MULTITREEm *MT)
{
  double         FinalMSE=0;
  DistanceInfo   *DI;
  CriterionInfo  *CI;
  Iterations     *Iter;
  int            ReturnValue;


  switch(Params->Distance)
  {
    case 1: // CIE DeltaE `76
      DI = diNew(TS, DT_User, &CIE76_D2D, &CIE76_D2M, &CIE76_M2D, &CIE76_M2M);
      break;
    case 2: // CIE DeltaE `96
      DI = diNew(TS, DT_User, &CIE94_D2D, &CIE94_D2M, &CIE94_M2D, &CIE94_M2M);
      break;
    default:
      DI = diNew(TS, ciCriterionDefaultDistance(CR_MSE), 0, 0, 0, 0);
  }

  switch(Params->CI)
  {
    case CR_MSE:
      CI = ciNew(TS, DI, CR_MSE, 0, 0, 0, 0, 0, 0);
      break;
    case CR_MAE:
      CI = ciNew(TS, DI, CR_MAE, 0, 0, 0, 0, 0, 0);
      break;
    case CR_FTest:
      CI = ciNew(TS, DI, CR_FTest, 0, 0, 0, 0, 0, 0);
      break;
  }
//WriteTrainingSet("ts0.ts", TS, 1);
//ReadTrainingSet("ts0.ts", TS);

if (Params->CutOff)
{
  if(Params->QuietLevel >= 2) PrintMessage("CutOff\n");
  CutOff1(Params, TS, MT);
  if(Params->QuietLevel >= 2) PrintMessage("Training set contains now %u cells.\n", BookSize(TS));
  PrintCodeBook(TS, 2);
}

//WriteTrainingSet("ts1.ts", TS, 1);
//ReadTrainingSet("ts1.ts", TS);
//PrintCodeBook(TS, Params->tweak);

  CreateNewCodebook(CB, Params->Colors, TS);
  GenerateRandomCodebook(TS, CB);          

  CreateNewPartitioning(PA, TS, Params->Colors);
  GenerateOptimalPartitioning(TS, CB, PA);

  Iter = itNew(I_Fixed, Params->IterRLS, 0, 0, 0);

  ReturnValue = RandomizedLocalSearch(TS, CB, PA, CI, &FinalMSE, Iter, Params->IterGLA, 1, Params->QuietLevel, 0, 0);
  
//WriteCodebook("cb0.cb", CB, 1);
//ReadCodebook("cb0.cb", CB);
//PrintCodeBook(CB, Params->tweak);

  itDelete(Iter);
  ciDelete(CI);
  diDelete(DI);

  return (FinalMSE);
}


/*-------------------------------------------------------------------*/
int GetColorIndex(TRAININGSET* TS, CODEBOOK* CB, PARTITIONING* PA, MULTITREEm* MMT, RGBPIXEL color, RGBPIXEL color1)
{
  int         CBindex=-1, TSindex=-1;
  int         i, r, g, b;
  VECTORTYPE  v;
  double      mindist, d;


  r = (int)RED(color);
  g = (int)GREEN(color);
  b = (int)BLUE(color);

  if (MMT->Children[r]!=NULL && MMT->Children[r]->Children[g]!=NULL && MMT->Children[r]->Children[g]->Children[b]!=NULL)
  {
    TSindex = MMT->Children[r]->Children[g]->Children[b]->index;
    if (TSindex!=-1 && PA!=NULL) CBindex=Map(PA,TSindex);
  }

  if(CBindex==-1)
  {
    v = Vector(CB, 0);
    CBindex=0;
    mindist = ColorDistance76(D65, (BYTE)RED(color),(BYTE)GREEN(color),(BYTE)BLUE(color), (BYTE)v[0],(BYTE)v[1],(BYTE)v[2]);

    for(i=0; i<BookSize(CB); i++)
    {
      v = Vector(CB,i);
      d = ColorDistance76(D65, (BYTE)RED(color),(BYTE)GREEN(color),(BYTE)BLUE(color), (BYTE)v[0],(BYTE)v[1],(BYTE)v[2]);
      if( d < mindist )
      {
        mindist  = d;
        CBindex = i;
      }
    }
    
    // CutOff 
    if (mindist<=DELAB) return (CBindex);
    else  // Threshold
    {
      v = Vector(CB, 0);
      CBindex=0;
      mindist = ColorDistance76(D65, (BYTE)RED(color),(BYTE)GREEN(color),(BYTE)BLUE(color), (BYTE)v[0],(BYTE)v[1],(BYTE)v[2]);

      for(i=0; i<BookSize(CB); i++)
      {
        v = Vector(CB,i);
        d = ColorDistance76(D65, (BYTE)RED(color),(BYTE)GREEN(color),(BYTE)BLUE(color), (BYTE)v[0],(BYTE)v[1],(BYTE)v[2]);
        if( d < mindist )
        {
          mindist  = d;
          CBindex = i;
        }
      }
    }
  }
  return (CBindex);
}


/*-------------------------------------------------------------------*/
RGBPIXEL GetColorByIndex(CODEBOOK* CB, int Index)
{
  VECTORTYPE  v;
  RGBPIXEL    result;

  v = Vector(CB, Index);
  result = RGB(v[0], v[1], v[2]);

  return(result);
}



/*-------------------------------------------------------------------*/
int BinToGrayCode(int value)
{
  int retvalue;
  
  retvalue = value ^ (value >> 1);

  return (retvalue);
}


int GrayToBinCode(int value)
{
  int retvalue;
  int mask = 0x80; 
  
  // Algorithm: bc[7]=gc[7]; for i=6..0: bc[i]=gc[i] XOR bc[i+1]
  
  retvalue = value & mask; // set highest bit, all the rest must be zero
  for (;mask!=0;mask>>=1)
  {
    if ((value ^ (retvalue>>1))&mask) // if XOR results in non-zero
      retvalue |= mask;                      // set appropriate bit
  }

  return (retvalue);
}


/*-------------------------------------------------------------------*/
YESNO CheckExt(char* FileName, char* Extention)
{
  char* ch;

  ch = strchr(FileName, '.');
  if (ch==NULL) return NO;

  ch ++;
  if(ch)
  {
    if (strcmp(ch, Extention)==0) return YES;
  }

  return NO;
}


/*-------------------------------------------------------------------*/
ULONG CalculateLayer(char *LayerName, int ContextSize)
{
  IMAGE   image;
  int     bit, i;
  ULONG   *Whites, *Total;
  ULONG   context, NContexts, it, x, y;
  double  *DynEntropy, sum;
  ULONG   CumDynEntropy, amount;


  NContexts = 2<<ContextSize;
  Whites = (ULONG*)allocate(sizeof(ULONG)*NContexts);
  if(Whites==NULL) return 0;
  Total  = (ULONG*)allocate(sizeof(ULONG)*NContexts);
  if(Total==NULL) return 0;
  DynEntropy = (double*)allocate(sizeof(double)*NContexts);
  if(DynEntropy==NULL) return 0;

  for(it=0;it<NContexts;it++)
  {
    Whites[it] = 0;
    Total[it] = 0;
    DynEntropy[it] = 0;
  }

  ImageInit(LayerName, &image, INPUT, NO, AUTOMATIC, BINMODE, DISK, ContextSize+10);

  SetProgressVisibility(1);
  OpenProgressWindow();
  for (y=1; y<=image.ImageSizeY; y++)
  {
    for (x=1; x<=image.ImageSizeX; x++)
    {
      bit = GetBitPixel(&image,x,y);
      
      context = 0;
      for(i=1; i<=ContextSize; i++)
      {
        context = (context<<1) + GetBitPixel(&image, x+DefTemplateX[i], y+DefTemplateY[i]);
      }  
      
      amount = (bit == BIN_WHITE ? Whites[context] : Total[context] - Whites[context]);
      if (Total[context] != 0 && amount != 0)
      DynEntropy[context] -= log((amount+DELTA) / (Total[context]+2*DELTA));

      Total[context]++;
      if (bit == BIN_WHITE) Whites[context]++;
    }
    UpdateProgressBar(y, image.ImageSizeY);
  }
  CloseProgressWindow();
  ImageDone(&image);

  sum = 0;
  for(it=0;it<NContexts;it++) sum +=DynEntropy[it];
  CumDynEntropy = (ULONG)sum;

  deallocate(Whites);
  deallocate(Total);
  deallocate(DynEntropy);

  return CumDynEntropy;
}
 

void CutOff0(SEGMPARAMS *Params, TRAININGSET *TS, MULTITREEm* MT)
{
  double       de;
  int          i, k=0;
  int          maxfreq, maxnode=1, currfreq, freq, cnt=0;
  TRAININGSET  TS1;
  VECTORTYPE   v, v1;
  double       L,A,B, L1,A1,B1, X,Y,Z;


  CreateNewCodebook(&TS1, BookSize(TS), TS);
  CopyCodebook(TS, &TS1);

  SetProgressVisibility(1);
  OpenProgressWindow();

  while(maxnode!=-1)
  {
    // Find MAX
    maxfreq=0;
    maxnode=-1;
    for(i=0; i<BookSize(&TS1); i++)
    {
      currfreq = Node(&TS1, i).freq;
      if (currfreq>maxfreq)
      {
        maxnode = i;
        maxfreq = currfreq;
      }
    }
    if (maxnode==-1) continue;

    // Find surroundings
    v = Vector(&TS1, maxnode);
    Node(&TS1, maxnode).freq = 0;

    for(i=0; i<BookSize(&TS1); i++)
    {
      v1 = Vector(&TS1, i);
      de = ColorDistance76(D65,(BYTE)v[0],(BYTE)v[1],(BYTE)v[2],(BYTE)v1[0],(BYTE)v1[1],(BYTE)v1[2]); 

      if (i!=maxnode && Node(&TS1, i).freq!=0 && de<DELAB)
      {
        //if(Params->QuietLevel>3) PrintMessage(" %d\t%f\tnode %d\n",maxnode,de,i);
        if(Params->QuietLevel>2)
        { 
          RGB_to_XYZ((BYTE)v[0],(BYTE)v[1],(BYTE)v[2], &X, &Y, &Z);
          XYZ_to_LAB(D65, X, Y, Z, &L, &A, &B);
          RGB_to_XYZ((BYTE)v1[0],(BYTE)v1[1],(BYTE)v1[2], &X, &Y, &Z);
          XYZ_to_LAB(D65, X, Y, Z, &L1, &A1, &B1);

          PrintMessage("%2.2f\t%2.2f\t%2.2f\t\t%d\t\t%2.2f\t%2.2f\t%2.2f\t%d\n",
                       L, A, B, Node(TS, maxnode).freq, L1, A1, B1, Node(&TS1, i).freq);
        }

        Node(&TS1, i).freq = 0;
        freq = Node(TS, i).freq;
        
        if (Node(TS, i).freq>=4) Node(TS, i).freq /= 4;
        else Node(TS, i).freq = 1;

        freq -= Node(TS, i).freq;
        Node(TS, maxnode).freq += freq;
        cnt++;
      }
    }
    if(Params->QuietLevel >= 2) UpdateProgressBar(cnt, BookSize(&TS1));
  }
  CloseProgressWindow();
}


void CutOff1(SEGMPARAMS *Params, TRAININGSET *TS, MULTITREEm* MT)
{
  double       de;
  int          i, k=0, r, g, b;
  int          maxfreq, maxnode=1, currfreq, freq, cnt=0;
  TRAININGSET  TS1;
  VECTORTYPE   v, v1;
  double       L,A,B, L1,A1,B1, X,Y,Z;


  CreateNewCodebook(&TS1, BookSize(TS), TS);
  CopyCodebook(TS, &TS1);

  SetProgressVisibility(1);
  OpenProgressWindow();

  while(maxnode!=-1)
  {
    // Find MAX
    maxfreq=0;
    maxnode=-1;
    for(i=0; i<BookSize(&TS1); i++)
    {
      currfreq = Node(&TS1, i).freq;
      if (currfreq>maxfreq)
      {
        maxnode = i;
        maxfreq = currfreq;
      }
    }
    if (maxnode==-1) continue;

    // Find surroundings
    v = Vector(&TS1, maxnode);
    Node(&TS1, maxnode).freq = 0;

    for(i=0; i<BookSize(&TS1); i++)
    {
      if (i!=maxnode && Node(&TS1, i).freq!=0)
      {
        v1 = Vector(&TS1, i);
        r = v1[0];
        g = v1[1];
        b = v1[2];

        de = ColorDistance76(D65,(BYTE)v[0],(BYTE)v[1],(BYTE)v[2],(BYTE)r,(BYTE)g,(BYTE)b); 
        if (de<DELAB)
        {
          if(Params->QuietLevel>2)
          {
            RGB_to_XYZ((BYTE)v[0],(BYTE)v[1],(BYTE)v[2], &X, &Y, &Z);
            XYZ_to_LAB(D65, X, Y, Z, &L, &A, &B);
            RGB_to_XYZ((BYTE)r,(BYTE)g,(BYTE)b, &X, &Y, &Z);
            XYZ_to_LAB(D65, X, Y, Z, &L1, &A1, &B1);

            PrintMessage("%2.2f\t%2.2f\t%2.2f\t\t%d\t\t%2.2f\t%2.2f\t%2.2f\t%d\n",
                         L, A, B, Node(TS, maxnode).freq, L1, A1, B1, Node(TS, i).freq);
          }
          freq = Node(TS, i).freq;
          Node(&TS1, i).freq = 0;
          Node(TS, i).freq = 0;
          Node(TS, maxnode).freq += freq;

          if (MT->Children[r]!=NULL && MT->Children[r]->Children[g]!=NULL && MT->Children[r]->Children[g]->Children[b]!=NULL)
          {
            MT->Children[r]->Children[g]->Children[b]->index=-1;
          }
          cnt++;
        }
      }
    }
    if(Params->QuietLevel >= 2) UpdateProgressBar(cnt, BookSize(&TS1));
  }
  CloseProgressWindow();

  cnt=0;
  for(i=0; i<BookSize(TS); i++)
  {
    if(Node(TS, i).freq!=0) { cnt++; }
  }

  FreeCodebook(&TS1);
  CreateNewCodebook(&TS1, cnt, TS);
  CopyCodebookHeader(TS, &TS1);
  BookSize(&TS1) = cnt;
  k=0;
  for (i=0; i<BookSize(TS); i++)
  {
    if(Node(TS, i).freq!=0)
    {
      CopyNode(&Node(TS,i), &Node(&TS1,k), VectorSize(TS));
      v = Vector(TS, i);
      r = v[0];
      g = v[1];
      b = v[2];
      if (MT->Children[r]!=NULL && MT->Children[r]->Children[g]!=NULL && MT->Children[r]->Children[g]->Children[b]!=NULL)
      {
        MT->Children[r]->Children[g]->Children[b]->index=k;
      }
      k++;
    }
  }

  FreeCodebook(TS);
  CreateNewCodebook(TS, BookSize(&TS1), &TS1);
  CopyCodebook(&TS1, TS);

  FreeCodebook(&TS1);
}


/* ----------------------------------------------------------------- */
/* -- New node                                                    -- */
/* ----------------------------------------------------------------- */
MULTITREEm *NewMTreeNode()
{
  MULTITREEm *CT;

  CT=(MULTITREEm*)allocate(sizeof(MULTITREEm));
  CT->Children = NULL;
  CT->index = 0;
  CT->freq = 0;

  NodesAdded++;
  return(CT);
}


/* ----------------------------------------------------------------- */
/* -- Tree construction                                           -- */
/* ----------------------------------------------------------------- */
ULONG ConstructColorTree(MULTITREEm **MT, IMAGE* Image, IMAGE* Layer, int QuietLevel, int tweak, int thr)
{
  MULTITREEm    *T;
  ULONG        y, x, i, j;
  RGBPIXEL     pixel, LPixel;
  VECTORTYPE   v;


  if(tweak) v = CreateEmptyVector(4);
  else v = CreateEmptyVector(3);

  (*MT) = NewMTreeNode();
  Colors=0;
  NodesAdded=0;

  if(QuietLevel>2)
  {
    SetProgressVisibility(1);
    OpenProgressWindow();
  }
  for( y=1; y<=Image->ImageSizeY; y++)
  {
    for( x=1; x<=Image->ImageSizeX; x++)
    {
      T=(*MT);
      pixel = GetRGBPixelEx(Image, x, y);
      if (Layer!=NULL) LPixel=GetGrayPixel(Layer, x, y);
      else LPixel=0; 
      if (LPixel==0 || LPixel>(ULONG)thr)         // DELAB=2.3   LPixel=0 <- for uniform filled areas
      {
        v[0] = RED(pixel);
        v[1] = GREEN(pixel);
        v[2] = BLUE(pixel);
        if(tweak) v[3] = GetExtraProperty(Image, x, y);
     
        for (i=0;i<3;i++)
        {
          if (T->Children==NULL)
          {
            T->Children=(MULTITREEm**)allocate(sizeof(MULTITREEm*)*256);
            assert(T->Children!=NULL);
            for(j=0;j<256;j++) T->Children[j]=NULL;
          }
          
          if (T->Children[v[i]]==NULL)
          {
            T->Children[v[i]]=NewMTreeNode();
            if (i==2) Colors++;
          }
          T=T->Children[v[i]];
        }
        (T->freq)++;
      }
    }
    if(QuietLevel>2) UpdateProgressBar(y, Image->ImageSizeY);
  }
  if(QuietLevel>2) CloseProgressWindow();

  FreeVector(v);
  return (Colors);
}


/*-------------------------------------------------------------------*/
int GetExtraProperty(IMAGE* image, ULONG x, ULONG y)
{ 
  RGBPIXEL     pl;
  ULONG        i;
  double       L, minL, maxL;

  
  pl = GetRGBPixelEx(image, x+Block3x3X[1], y+Block3x3Y[1]);
  L  = (float)(RED(pl))*0.3 + (float)(GREEN(pl))*0.586 + (float)(BLUE(pl))*0.115;
  minL=L;
  maxL=L;
  
  for(i=2; i<=9; i++)
  {
    pl = GetRGBPixelEx(image, x+Block3x3X[i], y+Block3x3Y[i]);
    L  = (float)(RED(pl))*0.3 + (float)(GREEN(pl))*0.586 + (float)(BLUE(pl))*0.115;
    if (L<minL) minL=L;
    if (L>maxL) maxL=L;
  }  
  return ((int)(maxL-minL));
}

/*-------------------------------------------------------------------*/
ULONG ReadRAWToColorTree(MULTITREEm **MT, BYTE *Data, ULONG XSize, ULONG YSize)
{
  VECTORTYPE  v;
  ULONG       x, y;
  BYTE        *tmp;
  MULTITREEm   *T;
  int         i, j;


  (*MT) = NewMTreeNode();
  Colors=0;
  NodesAdded=0;
  tmp = Data;
  v=CreateEmptyVector(3);

  for(y=1; y<=YSize; y++)
  {
    for(x=1; x<=XSize; x++)
    {
      T=(*MT);

      v[0] = tmp[0];
      v[1] = tmp[1];
      v[2] = tmp[2];
      
      for (i=0;i<3;i++)
      {
        if (T->Children==NULL)
        {
          T->Children=(MULTITREEm**)allocate(sizeof(MULTITREEm*)*256);
          assert(T->Children!=NULL);
          for(j=0;j<256;j++) T->Children[j]=NULL;
        }
          
        if (T->Children[v[i]]==NULL)
        {
          T->Children[v[i]]=NewMTreeNode();
          if (i==2) Colors++;
        }
        T=T->Children[v[i]];
      }
      (T->freq)++;
      tmp +=3;
    }
  }
  FreeVector(v);
  return (Colors);
}



/* ----------------------------------------------------------------- */
void DeallocateMTree(MULTITREEm *MT)
{
  int i;
  
  
  if (MT->Children!=NULL)
  {
    for(i=0;i<256;i++)
    {
      if (MT->Children[i]!=NULL)
      {
        DeallocateMTree(MT->Children[i]);
      }
    }
  }
  deallocate(MT->Children);
  deallocate(MT);
}