/***********************************************
**
** Filename: HistoDataPolar.h
**   
** Basic functions for HistoDataPolar object
**
**
** Date  
**    27.02.2003   
**    01.03.2004  Polar Quantization   
**    19.04.2004  Polar  Class 
**
** Written by: 
**    Alexander Kolesnikov 
**
** Email:
**   koles@cs.joensuu.fi
**
**
**  Copyright (c) 2003,2004 Alexander Kolesnikov 
**
**********************************************/


#ifndef HISTODATAPOLAR_H
#define HISTODATAPOLAR_H

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <stddef.h>
#include <string.h>
#include <time.h>

//typedef unsigned long int ULONG;

#include "VectSTD.h"
#include "HistoData.h"
#include "SearchBasic.h"

extern "C"
{
	#include "../Modules/cb.h"
}

class CHistoData;

class CHistoDataPolar
{
public:

  /**********************************************/
  /* Constructor, Destructor, Copy method       */
  /**********************************************/
   
	CHistoDataPolar();
	
	CHistoDataPolar(double *pR, double *pPhi, long int size); 
	CHistoDataPolar(double *pR, double *pPhi, 
				 unsigned long int *pHisto, long int size);

	CHistoDataPolar(CHistoDataPolar& HistoData);
	CHistoDataPolar(DPolyline *pPoly); 
    CHistoDataPolar(long int size);
    CHistoDataPolar(DArray *pR, DArray *pPhi); 


	virtual ~CHistoDataPolar();

	/**********************************************/
    // Heuristic approach
	/**********************************************/
    int Uniform(int Mr, int Mphi, double *E2);
	int MaxLloyd(ULONG Mr, ULONG Mphi, double *E2);
	int MaxLloyd(ULONG M, ULONG *Mr, ULONG *Mphi, double *E2);

    /////////////////////////////////////////////////
    // Load data
    ////////////////////////////////////////////////
    //int Load(char* filename);
    //FILE*  OpenForReadingAndCheckLabel(char* filename, char *label);

   /////////////////////////////////////
	// G E T
	/////////////////////////////////////
	//int GetPartitionX(DArray ** ppPart);
    double* GetThresholdsR();
	double* GetThresholdsPhi();

	long int GetClustersNumberR();
	long int GetClustersNumberPhi();

	double* GetMeansR(unsigned long Mr);
	double* GetMeansPhi(unsigned long Mphi);

	double* GetVariancesR();
	double* GetVariancesPhi();

	/*********************************************/
    // Type data
	/*********************************************/
    void   Type();
    int    TypeHistoData();
	int    TypeClustersList();

	int    Copy(CHistoDataPolar* pHisto_dst);


	// Statsical characteristics
    void    GetThresholds(double* pThr);
	int     CalcMoments();
    int     SortPointsByPhase();
    void    SortDataByPhase();
	void    SortDataByRadius();

//	void    SwapPolarPoints(DPolarPoint *p1, DPolarPoint *p2);
    //int     CalcClustersList();
	int     CalcVariancePhase(double *dErr);

	/*********************************************/
	// Memory management
	/*********************************************/
    int    InitMemory(ULONG size);
    void   FreeMemory(); 
    int    ReallocateMemory(ULONG size); 
    
    /**********************************************/
	// Dynamic Programming Aapproach
	/**********************************************/
    int DP_Product_One(ULONG Mr, ULONG Mphi, double *E2); 
	int DP_Product_Many(ULONG M, ULONG *Mr, ULONG *Mphi, double *E2); 

    int DP_MSE_PolarQ_1(ULONG M, int nType, double *error_value); 

	int SPQ(ULONG Mr, ULONG Mphi, double *error_value); 
	int SPQ(ULONG M, ULONG *Mr, ULONG *Mphi, double *error_value); 

    int CalcFunctional(double *value);    

    /*********************************************/
    // Dynamic Programming approach: Private
	/*********************************************/
    int DP_MSE_min(Bounds *pBnd, DTable  *pD2, NTable *pNodes);
    int DP_Phase_MSE_min(ULONG Mphi, int nType, double *E2);
    int DP_Phase_search(Bounds *pBnd, DTable  *pD2, NTable *pNodes); 

	// Unconstrained Polar Quantization
	int UPQ(ULONG M, ULONG Mr, double *E2, ULONG *phi_c, int mode);
	int UPQ_Full(ULONG M, ULONG Mr, double *E2, CODEBOOK *CB);

    int PostOptimization(DArray **ppE2, NArray *pM_min, NArray *pM_num, NArray *pM_max, 
					 long int rings_num, double *dE2);

	double MakeConfigurationDP(int CurrentMaxNumber, DArray **Errors, NArray *Config,
								int NumberOfRings);

    int ConstructRings(ULONG Mr, DArray **ppPhi);

	// Sorting routines
	void heapSortRadius();
	void siftDownRadius(ULONG root, ULONG bottom);

	void heapSortPhase();
	void siftDownPhase(ULONG root, ULONG bottom);

// private:
	// Input data
    long int     m_nSize;           // Number of points     
	double       *m_pR;            
	double       *m_pPhi;          
	
	// Data for Points (Radius,Phase)
//	DPolarPoint  *m_pPoints;
	ULONG        *m_pHisto;          
	double       *m_pProby;          

    double       *m_pE2;         

	// Objects 
	CHistoData  *m_pHData_R;
	CHistoData  *m_pHData_Phi;

     // Moments
	double     *m_pPhaseMomentZero;  // Cumulative PolarMoment Zero:  p*1     ; 
    double     *m_pPhaseMomentOne;   // Cumulative PolarMoment One :  p*phi   ;
    double     *m_pPhaseMomentTwo;   // Cumulative PolarMoment Two :  p*phi^2 ;

	double     *m_pRadiusMomentZero; // Cumulative PolarMoment Zero:  p*1     ; 
    double     *m_pRadiusMomentOne;  // Cumulative PolarMoment One :  p*phi   ;
    double     *m_pRadiusMomentTwo;  // Cumulative PolarMoment Two :  p*phi^2 ;

	double     *m_pPolarMomentZero; // Cumulative PolarMoment Zero:  p*1     * r^2; 
    double     *m_pPolarMomentOne;  // Cumulative PolarMoment One :  p*phi   * r^2;
    double     *m_pPolarMomentTwo;  // Cumulative PolarMoment Two :  p*phi^2 * r^2;
};


#endif /* HISTODATA_H */
